# Copyright (c) 2025 Bytedance Ltd. and/or its affiliates
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional, TypeVar, Union

from pydantic import BaseModel

from arkitect.core.errors import ArkError, Error


class Request(BaseModel):
    stream: bool = False


class Response(BaseModel):
    error: Optional[Union[Error, ArkError]] = None


RequestType = TypeVar("RequestType", bound=Request, contravariant=True)
ResponseType = TypeVar("ResponseType", bound=Response, covariant=True)


class Context(BaseModel):
    request_id: str = ""
    client_request_id: str = ""
    account_id: str = ""
    resource_id: str = ""
    resource_type: str = ""



# Copyright (c) 2025 Bytedance Ltd. and/or its affiliates
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from arkitect.core.component.tts.base import AsyncBaseTTSClient, TTSResponseChunk
from arkitect.core.component.tts.bot_util import create_bot_audio_responses
from arkitect.core.component.tts.model import AudioParams, ConnectionParams, TextRequest
from arkitect.core.component.tts.tts_client import AsyncTTSClient

__all__ = [
    "AsyncBaseTTSClient",
    "TTSResponseChunk",
    "AsyncTTSClient",
    "ConnectionParams",
    "AudioParams",
    "TextRequest",
    "create_bot_audio_responses",
]



# Copyright (c) 2025 Bytedance Ltd. and/or its affiliates
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from arkitect.core.component.llm.model import (
    ArkChatRequest,
    ArkMessage,
    ChatCompletionMessageImageUrlPart,
    ChatCompletionMessageImageUrlPartImageUrl,
)
from arkitect.utils import dump_json_truncate


def test_dump_json():
    obj = ArkChatRequest(
        messages=[
            ArkMessage(
                role="user",
                content=[
                    ChatCompletionMessageImageUrlPart(
                        type="image_url",
                        image_url=ChatCompletionMessageImageUrlPartImageUrl(
                            url="fake-b64-url",
                        ),
                    )
                ],
            )
        ],
        model="fake-model",
    )
    obj_copied = dump_json_truncate(obj, 1)
    assert obj_copied["messages"][0]["content"][0]["image_url"]["url"] == "f"
    obj_copied = dump_json_truncate(obj, 100)
    assert obj_copied["messages"][0]["content"][0]["image_url"]["url"] == "fake-b64-url"



# Copyright (c) 2025 Bytedance Ltd. and/or its affiliates
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import gzip
import json
from typing import Union

from arkitect.core.component.tts.constants import (
    GZIP,
    INT_SIZE,
    JSON,
    NO_SERIALIZATION,
    WITH_EVENT,
    EventConnectionFailed,
    EventConnectionFinished,
    EventConnectionStarted,
    EventFinishConnection,
    EventSessionFinished,
    EventStartConnection,
)
from arkitect.core.component.tts.model import ResponseEvent


def contain_event(flags: int) -> bool:
    return flags == WITH_EVENT


def parse_response(res: Union[str, bytes]) -> ResponseEvent:
    """
    protocol_version(4 bits), header_size(4 bits),
    message_type(4 bits), message_type_specific_flags(4 bits)
    serialization_method(4 bits) message_compression(4 bits)
    reserved （8bits) reserved(8bits)
    header_extensions header extension(size = 8 * 4 * (header_size - 1) )
    payload same as http body
    """
    if isinstance(res, str):
        res = res.encode("utf-8")
    header_size = res[0] & 0x0F
    message_type_specific_flags = res[1] & 0x0F
    serialization_method = res[2] >> 4
    message_compression = res[2] & 0x0F
    ptr = header_size * 4
    result = ResponseEvent()
    if contain_event(message_type_specific_flags):
        event = int.from_bytes(res[ptr : ptr + INT_SIZE], "big", signed=True)
        ptr += 4
        result.event = event
        if event == EventSessionFinished:
            result.session_finished = True
        if event not in [
            EventStartConnection,
            EventFinishConnection,
            EventConnectionStarted,
            EventConnectionFailed,
            EventConnectionFinished,
        ]:
            session_id_len = int.from_bytes(
                res[ptr : ptr + INT_SIZE], "big", signed=False
            )
            ptr += 4
            session_id = res[ptr : ptr + session_id_len]
            ptr += session_id_len
            result.session_id = session_id.decode("utf-8")
        if event in [
            EventConnectionStarted,
            EventConnectionFailed,
            EventConnectionFinished,
        ]:
            connection_id_len = int.from_bytes(
                res[ptr : ptr + INT_SIZE], "big", signed=False
            )
            ptr += 4
            connection_id = res[ptr : ptr + connection_id_len]
            ptr += connection_id_len
            result.connection_id = connection_id.decode("utf-8")
    payload_size = int.from_bytes(res[ptr : ptr + INT_SIZE], "big", signed=True)
    payload = res[ptr + INT_SIZE :]
    payload_msg = payload

    if message_compression == GZIP:
        payload_msg = gzip.decompress(payload_msg)
    if serialization_method == JSON:
        result.payload_msg = json.loads(payload_msg)
    elif serialization_method == NO_SERIALIZATION:
        result.audio_only = True
        result.audio = payload_msg
    result.payload_size = payload_size
    return result



# Copyright (c) 2025 Bytedance Ltd. and/or its affiliates
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import abstractmethod
from typing import Any, Optional, TypeVar

from langchain_core.output_parsers import BaseOutputParser
from langchain_core.prompts import BasePromptTemplate
from pydantic.v1 import BaseModel, Field
from volcenginesdkarkruntime import AsyncArk

from arkitect.core.client import default_ark_client

T = TypeVar("T")


class BaseLanguageModel(BaseModel):
    endpoint_id: str
    client: AsyncArk = Field(default_factory=default_ark_client)
    template: Optional[BasePromptTemplate] = None
    output_parser: Optional[BaseOutputParser] = None

    class Config:
        """Configuration for this pydantic object."""

        arbitrary_types_allowed = True

    @abstractmethod
    def _run(self, *args: Any, **kwargs: Any) -> Any:
        raise NotImplementedError

    @abstractmethod
    async def _arun(self, *args: Any, **kwargs: Any) -> Any:
        raise NotImplementedError



# Copyright (c) 2025 Bytedance Ltd. and/or its affiliates
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .errorcode import ArkError, Error
from .exceptions import (
    FALLBACK_EXCEPTIONS,
    AccountOverdueError,
    APIException,
    APITimeoutError,
    InternalServiceError,
    InvalidParameter,
    MissingParameter,
    RateLimitExceeded,
    ResourceNotFound,
    SensitiveContentDetected,
    ServerOverloaded,
    parse_pydantic_error,
)

__all__ = [
    "APIException",
    "InternalServiceError",
    "InvalidParameter",
    "MissingParameter",
    "RateLimitExceeded",
    "ServerOverloaded",
    "SensitiveContentDetected",
    "AccountOverdueError",
    "ResourceNotFound",
    "APITimeoutError",
    "FALLBACK_EXCEPTIONS",
    "Error",
    "ArkError",
    "parse_pydantic_error",
]



# Copyright (c) 2025 Bytedance Ltd. and/or its affiliates
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import random
from datetime import datetime

time_fmt = "%Y%m%d%H%M%S"


def gen_log_id() -> str:
    """
    Generates a unique log ID.

    This function creates a log ID by combining
    the current date and time with a random number.
    The date and time are formatted as YYYYMMDDHHMMSS,
    and the random number is a 20-digit hexadecimal number.
    """
    return datetime.now().strftime(time_fmt) + format(
        random.randint(0, 2**64 - 1), "020X"
    )



# Copyright (c) 2025 Bytedance Ltd. and/or its affiliates
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import io
import json
import logging
import unittest
from os import linesep
from typing import IO, List, Optional

from opentelemetry import trace
from opentelemetry.sdk.trace import TracerProvider
from opentelemetry.sdk.trace.export import ConsoleSpanExporter, SimpleSpanProcessor

from arkitect.telemetry.trace.wrapper import _update_kwargs, task


@task()
def test_sync(num: Optional[int] = None) -> int:
    if num is None:
        num = 0
    if num >= 2:
        return num
    return test_sync(num + 1)


@task()
async def test_async(num: Optional[int] = None) -> int:
    if num is None:
        num = 0
    if num >= 2:
        return num
    return await test_async(num + 1)


def setup_tracing(buffer: IO) -> None:
    trace.set_tracer_provider(TracerProvider())
    exporter = ConsoleSpanExporter(
        service_name="test_task",
        out=buffer,
        formatter=lambda span: span.to_json(indent=None) + linesep,
    )
    trace.get_tracer_provider().add_span_processor(SimpleSpanProcessor(exporter))


class TestTask(unittest.TestCase):
    def setUp(self):
        self.f = io.StringIO()
        setup_tracing(self.f)

    def test_update_kwargs(self):
        def func(request, config, extra=None):
            pass

        args = ["1", "2"]
        kwargs = {"a": "3", "b": "4"}
        inpt = _update_kwargs(args, kwargs, func)
        self.assertEqual(inpt.get("request"), "1")
        self.assertEqual(inpt.get("config"), "2")
        self.assertEqual(inpt.get("a"), "3")
        self.assertEqual(inpt.get("b"), "4")
        self.assertEqual(kwargs.get("request"), None)
        self.assertEqual(kwargs.get("config"), None)

        args = ["1", "2"]
        kwargs = None
        inpt = _update_kwargs(args, kwargs, func)
        self.assertEqual(inpt.get("request"), "1")
        self.assertEqual(inpt.get("config"), "2")

        args = None
        kwargs = {"a": "3", "b": "4"}
        inpt = _update_kwargs(args, kwargs, func)
        self.assertEqual(inpt.get("a"), "3")
        self.assertEqual(inpt.get("b"), "4")

        args = None
        kwargs = None
        inpt = _update_kwargs(args, kwargs, func)
        self.assertEqual(inpt, {})

    def test_sync_nested(self):
        test_sync()
        spans = []
        for line in self.f.getvalue().splitlines():
            span = json.loads(line)
            spans.append(span)
        self.assertEqual(len(spans), 3)
        self.assertEqual(spans[2]["parent_id"], None)
        self.assertEqual(spans[1]["parent_id"], spans[2]["context"]["span_id"])
        self.assertEqual(spans[0]["parent_id"], spans[1]["context"]["span_id"])

    def test_async_nested(self):
        loop = asyncio.get_event_loop()
        loop.run_until_complete(test_async())
        spans = []
        for line in self.f.getvalue().splitlines():
            span = json.loads(line)
            spans.append(span)
        self.assertEqual(len(spans), 3)
        self.assertEqual(spans[2]["parent_id"], None)
        self.assertEqual(spans[1]["parent_id"], spans[2]["context"]["span_id"])
        self.assertEqual(spans[0]["parent_id"], spans[1]["context"]["span_id"])

    def test_async_sync_mixed(self):
        loop = asyncio.get_event_loop()
        loop.run_until_complete(test_async(test_sync()))
        spans = []
        for line in self.f.getvalue().splitlines():
            span = json.loads(line)
            spans.append(span)
        self.assertEqual(len(spans), 4)
        self.assertEqual(spans[3]["parent_id"], None)
        self.assertEqual(spans[2]["parent_id"], None)
        self.assertEqual(spans[1]["parent_id"], spans[2]["context"]["span_id"])
        self.assertEqual(spans[0]["parent_id"], spans[1]["context"]["span_id"])

    def test_simulated(self):
        @task()
        async def chat() -> List[str]:
            return ["hello", "world"]

        @task()
        async def main(_messages: List[str]) -> List[str]:
            return await chat()

        @task()
        async def assistant_chat():
            await main(["Who are you"])

        loop = asyncio.get_event_loop()
        loop.run_until_complete(assistant_chat())
        spans = []
        for line in self.f.getvalue().splitlines():
            span = json.loads(line)
            spans.append(span)
        self.assertEqual(len(spans), 3)
        self.assertEqual(spans[2]["parent_id"], None)
        self.assertEqual(spans[1]["parent_id"], spans[2]["context"]["span_id"])
        self.assertEqual(spans[0]["parent_id"], spans[1]["context"]["span_id"])


if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG)
    unittest.main()



# Copyright (c) 2025 Bytedance Ltd. and/or its affiliates
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import gzip
import struct
import uuid
from abc import ABC, abstractmethod
from typing import Any, AsyncIterable, Optional

import websockets

from arkitect.core.component.asr.constants import DEFAULT_ASR_AUDIO
from arkitect.core.component.asr.model import (
    ASRAudio,
    ASRAudioInfoRsp,
    ASRAudioOnlyRequest,
    ASRFullClientRequest,
    ASRFullServerResponse,
    ASRRequest,
    ASRResult,
    ASRUser,
)
from arkitect.telemetry.logger import INFO
from arkitect.telemetry.trace import task
from arkitect.utils.binary_protocol import (
    AUDIO_ONLY_REQUEST,
    NO_SEQUENCE,
    POS_SEQUENCE,
    generate_before_payload,
    generate_header,
    parse_response,
)

__all__ = ["BaseAsyncASRClient", "AsyncASRClient"]


class BaseAsyncASRClient(ABC):
    @abstractmethod
    async def init(self) -> None:
        pass

    @abstractmethod
    async def reset_conn(self) -> None:
        pass

    @abstractmethod
    async def stream_asr(
        self, stream_audio: AsyncIterable[bytes], **kwargs: Any
    ) -> AsyncIterable[ASRFullServerResponse]:
        pass

    @abstractmethod
    async def close(self) -> None:
        pass


class AsyncASRClient(BaseAsyncASRClient, ABC):
    def __init__(
        self,
        access_key: str,
        app_key: str,
        api_resource_id: str = "volc.bigasr.sauc.duration",
        conn_id: str = str(uuid.uuid4()),
        log_id: str = str(uuid.uuid4()),
        base_url: str = "wss://openspeech.bytedance.com/api/v3/sauc/bigmodel",
        audio_format: ASRAudio = DEFAULT_ASR_AUDIO,
        user: Optional[ASRUser] = None,
        model_name: str = "bigmodel",
    ):
        self.api_resource_id = api_resource_id
        self.access_key = access_key
        self.app_key = app_key
        self.conn_id = conn_id
        self.log_id = log_id

        self.base_url = base_url
        self.audio_format = audio_format
        self.user = user
        self.model_name = model_name

        self.conn: Optional[websockets.WebSocketClientProtocol] = None
        self.session_id: Optional[str] = None
        self.inited = False

    async def init(self) -> None:
        if self.inited:
            return
        # init ws conn
        headers = {
            "X-Api-Resource-Id": self.api_resource_id,
            "X-Api-Access-Key": self.access_key,
            "X-Api-App-Key": self.app_key,
            "X-Api-Request-Id": self.log_id,
        }

        self.conn = await websockets.connect(self.base_url, extra_headers=headers)
        INFO(f"Connected to {self.base_url}, log_id: {self.log_id}")

        # send init response
        init_response = await self._send_full_client_request(
            ASRFullClientRequest(
                user=self.user,
                audio=self.audio_format,
                request=ASRRequest(model_name=self.model_name),
            )
        )

        INFO(f"Inited asr client: {init_response}")
        self.inited = True

    async def reset_conn(self) -> None:
        if self.conn:
            await self.conn.close()
            self.inited = False
        await self.init()
        INFO("Reset ASR Connection")

    async def stream_asr(
        self, stream_audio: AsyncIterable[bytes], **kwargs: Any
    ) -> AsyncIterable[Optional[ASRFullServerResponse]]:
        """
        Streams audio data to the ASR server and yields responses as they are received.
        """

        @task()
        async def send_audio_task(
            audio_stream: AsyncIterable[bytes],
        ) -> None:
            # i = 1
            async for data in audio_stream:
                await self._send_audio(
                    audio_only_request=ASRAudioOnlyRequest(
                        last_package=False,
                        seq=0,
                        audio=data,
                    )
                )
                # i += 1

        @task()
        async def receive_response_task() -> AsyncIterable[ASRFullServerResponse]:
            while True:
                if not self.inited:
                    INFO("ASR client is disconnected, skip the audio input.")
                    await asyncio.sleep(1)
                    continue
                response = await self._receive_response()
                INFO(f"Received asr server response: {response}")
                yield response

        t = asyncio.create_task(send_audio_task(audio_stream=stream_audio))

        async for rsp in receive_response_task():
            yield rsp

        await t

    async def close(self) -> None:
        if self.conn is not None:
            await self.conn.close()
            self.conn = None
        self.inited = False

    @task()
    async def _send_full_client_request(
        self, full_client_request: ASRFullClientRequest
    ) -> ASRFullServerResponse:
        payload_bytes = str.encode(
            full_client_request.model_dump_json(exclude_none=True, exclude_unset=True)
        )
        payload_bytes = gzip.compress(payload_bytes)
        full_client_bytes = bytearray(
            generate_header(message_type_specific_flags=POS_SEQUENCE)
        )
        full_client_bytes.extend(generate_before_payload(sequence=1))
        full_client_bytes.extend(
            (len(payload_bytes)).to_bytes(4, "big")
        )  # payload size(4 bytes)
        full_client_bytes.extend(payload_bytes)
        # payload

        await self.conn.send(full_client_bytes)
        res = await self.conn.recv()
        return ASRFullServerResponse(**parse_response(res))

    @task(watch_io=False)
    async def _send_audio(self, audio_only_request: ASRAudioOnlyRequest) -> None:
        if not self.conn or not self.inited:
            # connection is closed.
            INFO("ASR Conn is closed, will ignore the audio.")
            return
        payload_bytes = gzip.compress(audio_only_request.audio)
        audio_only_bytes = bytearray(
            generate_header(
                message_type=AUDIO_ONLY_REQUEST, message_type_specific_flags=NO_SEQUENCE
            )
        )
        audio_only_bytes.extend(struct.pack(">I", len(payload_bytes)))
        audio_only_bytes.extend(payload_bytes)  # payload
        await self.conn.send(audio_only_bytes)
        INFO(f"Sent Data INFO ASR SEVER data len={len(payload_bytes)}")

    async def _receive_response(self) -> Optional[ASRFullServerResponse]:
        if not self.conn:
            # connection is closed.
            return None
        res = await self.conn.recv()
        parsed_res = parse_response(res)
        # print(parsed_res)
        return ASRFullServerResponse(
            sequence=parsed_res.get("payload_sequence"),
            last_package=parsed_res.get("is_last_package", False),
            result=ASRResult(**parsed_res.get("payload_msg", {}).get("result", {})),
            audio=ASRAudioInfoRsp(
                **parsed_res.get("payload_msg", {}).get("audio_info", {})
            ),
        )



# Copyright (c) 2025 Bytedance Ltd. and/or its affiliates
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from typing import Any

from .common import LoggerName, Timer
from .logid import gen_log_id

__all__ = ["DEBUG", "INFO", "WARN", "ERROR", "Timer", "gen_log_id"]


def DEBUG(msg: str, *args: Any, **kwargs: Any) -> None:
    logging.getLogger(LoggerName.get()).debug(msg, stacklevel=2, *args, **kwargs)


def INFO(msg: str, *args: Any, **kwargs: Any) -> None:
    logging.getLogger(LoggerName.get()).info(msg, stacklevel=2, *args, **kwargs)


def WARN(msg: str, *args: Any, **kwargs: Any) -> None:
    logging.getLogger(LoggerName.get()).warn(msg, stacklevel=2, *args, **kwargs)


def ERROR(msg: str, *args: Any, **kwargs: Any) -> None:
    logging.getLogger(LoggerName.get()).error(msg, stacklevel=2, *args, **kwargs)



# Copyright (c) 2025 Bytedance Ltd. and/or its affiliates
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import AsyncIterator

from aiohttp import StreamReader


class AsyncSSEDecoder(object):
    """
    A class for decoding SSE response from a StreamReader.
    """

    def __init__(self, source: StreamReader) -> None:
        self.source = source

    async def _read(self) -> AsyncIterator[bytes]:
        data = b""
        async for chunk in self.source:
            for line in chunk.splitlines(True):
                data += line
                if data.endswith((b"\r\r", b"\n\n", b"\r\n\r\n")):
                    yield data
                    data = b""
        if data:
            yield data

    async def next(self) -> AsyncIterator[bytes]:
        """
        Decodes the next event from the SSE stream.
        """
        async for chunk in self._read():
            for line in chunk.splitlines():
                # skip comment
                if line.startswith(b":"):
                    continue

                if b":" in line:
                    field, value = line.split(b":", 1)
                else:
                    field, value = line, b""

                if field == b"data" and len(value) > 0:
                    yield value



# Copyright (c) 2025 Bytedance Ltd. and/or its affiliates
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from typing import Any, Callable, Dict, List, Optional, Tuple, Type

from arkitect.telemetry.trace import task
from arkitect.utils.common import Singleton


class Client(Singleton):
    """
    Base class for all clients.
    """

    pass


class ClientPool(Singleton):
    """
    A pool of clients that can be accessed by name.
    """

    _registry: Dict[str, Type[Client]] = {}
    clients: Dict[str, Client] = {}

    def __init__(
        self,
        clients: Dict[str, Tuple[Type[Client], Dict[str, Any]]],
    ) -> None:
        """
        Initialize the client pool with a dictionary of clients.

        :param clients: A dictionary of clients.
        """
        for name, (cls, config) in (clients or dict()).items():
            try:
                self.clients[name] = cls(**config)
            except Exception as e:
                logging.error(f"init client pool failed:{e}")
                continue

    def get_client_names(self) -> List[str]:
        """
        Get the names of all clients in the pool.

        :return: A list of client names.
        """
        return [name for name in self.clients.keys()]

    def get_client(self, name: str) -> Optional[Client]:
        """
        Get a client by name.

        :param name: The name of the client.
        :return: The client instance, or None if the client is not found.
        """
        return self.clients.get(name)

    @classmethod
    def register(
        cls, name: Optional[str] = None
    ) -> Callable[[Type[Client]], Type[Client]]:
        """
        Decorator to register a client class with the client pool.

        :param name: The name to register the class under.
                     If not provided, the class name will be used.
        :return: The decorated class.
        """

        def func(wrapped_cls: Type[Client]) -> Type[Client]:
            if name:
                cls._registry[name] = wrapped_cls
            else:
                cls._registry[wrapped_cls.__name__] = wrapped_cls

            return wrapped_cls

        return func

    @classmethod
    async def async_get_client(cls, name: str, config: Dict[str, Any]) -> Client:
        """
        Asynchronously get a client instance by name.
        If the client does not exist, create it.

        :param name: The name of the client.
        :param config: The configuration for the client.
        :return: The client instance.
        """
        if name not in cls.clients:
            cls.clients[name] = await cls.async_create_client(name, **config)

        return cls.clients[name]

    @classmethod
    async def async_create_client(cls, name: str, *args: Any, **kwargs: Any) -> Client:
        """
        Asynchronously create a client instance by name.

        :param name: The name of the client.
        :return: The client instance.
        """
        if name not in cls._registry:
            raise ValueError(f"Unknown client name: {name}")
        if not issubclass(cls._registry[name], Client):
            raise ValueError(f"{name} is not a subclass of client")
        client_cls = cls._registry[name]

        return await client_cls.get_instance_async(*args, **kwargs)


@task()
def get_client_pool(
    clients: Optional[Dict[str, Tuple[Type[Client], Any]]] = None,
) -> ClientPool:
    """
    Get the client pool instance.

    :param clients: A dictionary of clients.
    :return: The client pool instance.
    """
    return ClientPool.get_instance_sync(clients=clients)



# Copyright (c) 2025 Bytedance Ltd. and/or its affiliates
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc
from typing import Any, Callable, Iterable

from pydantic import BaseModel

from .model import RequestType, Response, ResponseType


class SyncRunner(BaseModel):
    invoke: Callable[[RequestType], Iterable[ResponseType]]

    class Config:
        """Configuration for this pydantic object."""

        arbitrary_types_allowed = True

    def __init__(
        self,
        runnable_func: Callable[[RequestType], Iterable[ResponseType]],
        **kwargs: Any,
    ):
        super().__init__(invoke=runnable_func, **kwargs)

    @abc.abstractmethod
    def run(self, request: RequestType) -> Response:
        pass

    @abc.abstractmethod
    def generate(self, request: RequestType) -> Iterable[str]:
        pass



# Copyright (c) 2025 Bytedance Ltd. and/or its affiliates
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import base64
import time
from typing import AsyncIterable, Union

import volcenginesdkarkruntime.types.chat.chat_completion_chunk as completion_chunk
from volcenginesdkarkruntime.types.chat.chat_completion import (
    ChatCompletionMessage,
    Choice,
)
from volcenginesdkarkruntime.types.chat.chat_completion_audio import ChatCompletionAudio

from arkitect.core.component.llm.model import (
    ArkChatCompletionChunk,
    ArkChatRequest,
    ArkChatResponse,
)
from arkitect.utils.context import get_client_reqid, get_reqid

from .base import TTSResponseChunk


def _get_first_chunk(request: ArkChatRequest) -> ArkChatCompletionChunk:
    return ArkChatCompletionChunk(
        id=get_reqid(),
        created=int(time.time()),
        model=request.model,
        choices=[
            completion_chunk.Choice(
                delta=completion_chunk.ChoiceDelta(
                    role="assistant",
                ),
                index=0,
            )
        ],
        object="chat.completion.chunk",
    )


def _get_normal_chunk(
    request: ArkChatRequest,
    chunk: TTSResponseChunk,
    first_audio_chunk: bool,
    first_transcript_chunk: bool,
    audio_id: str,
) -> ArkChatCompletionChunk:
    audio_chunk = {}
    if first_audio_chunk or first_transcript_chunk:
        audio_chunk["id"] = audio_id
    if chunk.audio:
        audio_chunk["data"] = base64.b64encode(chunk.audio).decode("utf-8")
    if chunk.transcript:
        audio_chunk["transcript"] = chunk.transcript
    return ArkChatCompletionChunk(
        id="1",
        created=int(time.time()),
        model=request.model,
        choices=[
            completion_chunk.Choice(
                delta=completion_chunk.ChoiceDelta(
                    audio=audio_chunk,
                ),
                index=0,
            )
        ],
        object="chat.completion.chunk",
    )


def _get_last_chunk(request: ArkChatRequest) -> ArkChatCompletionChunk:
    return ArkChatCompletionChunk(
        id=get_reqid(),
        created=int(time.time()),
        model=request.model,
        choices=[
            completion_chunk.Choice(
                delta=completion_chunk.ChoiceDelta(
                    audio={
                        "expires_at": int(time.time()),
                    },
                ),
                index=0,
            )
        ],
        object="chat.completion.chunk",
    )


async def _stream_bot_response_handler(
    tts_stream: AsyncIterable[TTSResponseChunk], request: ArkChatRequest, audio_id: str
) -> AsyncIterable[ArkChatCompletionChunk]:
    first_chunk = True
    first_audio_chunk = True
    first_transcript_chunk = True
    async for chunk in tts_stream:
        if first_chunk:
            yield _get_first_chunk(request)
            first_chunk = False
        if chunk.audio:
            yield _get_normal_chunk(
                request=request,
                chunk=chunk,
                first_audio_chunk=first_audio_chunk,
                first_transcript_chunk=first_transcript_chunk,
                audio_id=audio_id,
            )
            first_audio_chunk = False
        if chunk.transcript:
            yield _get_normal_chunk(
                request=request,
                chunk=chunk,
                first_audio_chunk=first_audio_chunk,
                first_transcript_chunk=first_transcript_chunk,
                audio_id=audio_id,
            )
            first_transcript_chunk = False

    yield _get_last_chunk(request)


async def _bot_response_handler(
    tts_stream: AsyncIterable[TTSResponseChunk], request: ArkChatRequest, audio_id: str
) -> AsyncIterable[ArkChatResponse]:
    audio_part: bytes = b""
    audio_transcript: str = ""
    async for chunk in tts_stream:
        if chunk.audio:
            audio_part += chunk.audio
        if chunk.transcript:
            audio_transcript += chunk.transcript
    yield ArkChatResponse(
        id=get_reqid(),
        model=request.model,
        choices=[
            Choice(
                index=0,
                message=ChatCompletionMessage(
                    role="assistant",
                    audio=ChatCompletionAudio(
                        id=audio_id,
                        expires_at=int(time.time()),
                        data=base64.b64encode(audio_part).decode("utf-8"),
                        transcript=audio_transcript,
                    ),
                ),
                finish_reason="stop",
            ),
        ],
        object="chat.completion",
        created=int(time.time()),
    )


async def create_bot_audio_responses(
    tts_stream: AsyncIterable[TTSResponseChunk], request: ArkChatRequest
) -> AsyncIterable[Union[ArkChatCompletionChunk, ArkChatResponse]]:
    audio_id = "audio_" + str(get_client_reqid())
    if request.stream:
        async for chunk in _stream_bot_response_handler(tts_stream, request, audio_id):
            yield chunk
    else:
        async for response in _bot_response_handler(tts_stream, request, audio_id):
            yield response



# Copyright (c) 2025 Bytedance Ltd. and/or its affiliates
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# type: ignore
import gzip
import json
from typing import Optional, Union

PROTOCOL_VERSION = 0b0001
DEFAULT_HEADER_SIZE = 0b0001

# Message Type:
FULL_CLIENT_REQUEST = 0b0001
AUDIO_ONLY_REQUEST = 0b0010
AUDIO_ONLY_SERVER = 0b1011
FULL_SERVER_RESPONSE = 0b1001
SERVER_ACK = 0b1011
SERVER_ERROR_RESPONSE = 0b1111

# Message Type Specific Flags
NO_SEQUENCE = 0b0000  # no check sequence
POS_SEQUENCE = 0b0001
NEG_SEQUENCE = 0b0010
NEG_WITH_SEQUENCE = 0b0011
NEG_SEQUENCE_1 = 0b0011

# Message Serialization
NO_SERIALIZATION = 0b0000
JSON = 0b0001

# Message Compression
NO_COMPRESSION = 0b0000
GZIP = 0b0001


def generate_header(
    message_type: int = FULL_CLIENT_REQUEST,
    message_type_specific_flags: int = NO_SEQUENCE,
    serial_method: int = JSON,
    compression_type: int = GZIP,
    reserved_data: int = 0x00,
) -> bytearray:
    """
    protocol_version(4 bits), header_size(4 bits),
    message_type(4 bits), message_type_specific_flags(4 bits)
    serialization_method(4 bits) message_compression(4 bits)
    reserved （8bits) 保留字段
    """
    header = bytearray()
    header_size = 1
    header.append((PROTOCOL_VERSION << 4) | header_size)
    header.append((message_type << 4) | message_type_specific_flags)
    header.append((serial_method << 4) | compression_type)
    header.append(reserved_data)
    return header


def generate_before_payload(sequence: int) -> bytearray:
    before_payload = bytearray()
    before_payload.extend(sequence.to_bytes(4, "big", signed=True))  # sequence
    return before_payload


def parse_response(res: bytes) -> dict:
    """
    protocol_version(4 bits), header_size(4 bits),
    message_type(4 bits), message_type_specific_flags(4 bits)
    serialization_method(4 bits) message_compression(4 bits)
    reserved （8bits) 保留字段
    header_extensions 扩展头(大小等于 8 * 4 * (header_size - 1) )
    payload 类似与http 请求体
    """
    # protocol_version = res[0] >> 4
    header_size = res[0] & 0x0F
    message_type = res[1] >> 4
    message_type_specific_flags = res[1] & 0x0F
    serialization_method = res[2] >> 4
    message_compression = res[2] & 0x0F
    # reserved = res[3]
    # header_extensions = res[4 : header_size * 4]
    payload = res[header_size * 4 :]
    result = {
        "is_last_package": False,
    }
    payload_msg = None
    payload_size = 0
    if message_type_specific_flags & 0x01:
        # receive frame with sequence
        seq = int.from_bytes(payload[:4], "big", signed=True)
        result["payload_sequence"] = seq
        payload = payload[4:]

    if message_type_specific_flags & 0x02:
        # receive last package
        result["is_last_package"] = True

    if message_type == FULL_SERVER_RESPONSE:
        payload_size = int.from_bytes(payload[:4], "big", signed=True)
        payload_msg = payload[4:]
    elif message_type == SERVER_ACK:
        seq = int.from_bytes(payload[:4], "big", signed=True)
        result["seq"] = seq
        if len(payload) >= 8:
            payload_size = int.from_bytes(payload[4:8], "big", signed=False)
            payload_msg = payload[8:]
    elif message_type == SERVER_ERROR_RESPONSE:
        code = int.from_bytes(payload[:4], "big", signed=False)
        result["code"] = code
        payload_size = int.from_bytes(payload[4:8], "big", signed=False)
        payload_msg = payload[8:]
    if payload_msg is None:
        return result
    if message_compression == GZIP:
        payload_msg = gzip.decompress(payload_msg)
    if serialization_method == JSON:
        payload_msg = json.loads(str(payload_msg, "utf-8"))
    elif serialization_method != NO_SERIALIZATION:
        payload_msg = str(payload_msg, "utf-8")
    result["payload_msg"] = payload_msg
    result["payload_size"] = payload_size
    return result


def parse_request(req: bytes) -> Optional[Union[bytes, dict]]:
    header_size = req[0] & 0x0F
    message_type = req[1] >> 4
    serialization_method = req[2] >> 4
    # message_compression = req[2] & 0x0F
    payload = req[header_size * 4 :]

    if message_type == FULL_CLIENT_REQUEST:
        # payload_size = int.from_bytes(payload[:4], "big", signed=True)
        payload_msg = payload[4:]
        if serialization_method == JSON:
            return json.loads(str(payload_msg, "utf-8"))
    elif message_type == AUDIO_ONLY_REQUEST:
        return payload

    return None



# Copyright (c) 2025 Bytedance Ltd. and/or its affiliates
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import json
import logging
from typing import Any, Dict, Optional, Union

from volcenginesdkarkruntime.types.chat import (
    ChatCompletion,
    ChatCompletionChunk,
)

from arkitect.core.component.tool import ArkToolResponse, ToolManifest
from arkitect.telemetry.trace import task
from arkitect.utils import dump_json_str

from .model import (
    ArkChatCompletionChunk,
    ArkChatRequest,
    ArkChatResponse,
    ArkMessage,
    FunctionCallMode,
)
from .utils import convert_response_message, transform_response


@task()
async def handle_function_call(
    request: ArkChatRequest,
    response: Union[
        ChatCompletionChunk, ChatCompletion, ArkChatCompletionChunk, ArkChatResponse
    ],
    functions: Optional[Dict[str, ToolManifest]] = None,
    function_call_mode: Optional[FunctionCallMode] = FunctionCallMode.SEQUENTIAL,
    **kwargs: Any,
) -> bool:
    """
    Handles function calls in a chat response.

    This function processes a chat response and checks if it contains a function call.
    If it does, it executes the function and appends the result to the request messages.

    Args:
        request : The original chat request.
        response : The chat response to process.
        functions : A dictionary of available functions.
        function_call_mode : The mode for handling function calls.
    """
    if response.choices[0].finish_reason != "tool_calls":
        return False

    response_message = (
        response.choices[0].delta
        if isinstance(response, ChatCompletionChunk)
        or isinstance(response, ArkChatCompletionChunk)
        else response.choices[0].message
    )
    tool_calls = response_message.tool_calls

    if not tool_calls or not functions:
        return False

    if function_call_mode and function_call_mode != FunctionCallMode.SEQUENTIAL:
        raise NotImplementedError("Only sequential function call mode is supported")

    request.messages.append(convert_response_message(response_message))

    function_calls = copy.deepcopy(tool_calls)
    for tool_call in function_calls:
        tool_name = tool_call.function.name

        tool = functions.get(tool_name)
        tool_response: ArkToolResponse = ArkToolResponse()
        if tool:
            parameters = json.loads(tool_call.function.arguments)
            tool_response = await tool.executor(parameters=parameters, **kwargs)

            logging.info(
                f"Function {tool_name} called with parameters:"
                + dump_json_str(parameters)
                + f" and response: {dump_json_str(tool_response)}"
            )
        else:
            logging.error(f"Function {tool_name} not found")

        request.messages.append(
            ArkMessage(
                role="tool",
                content=transform_response(tool_response.data),
                tool_call_id=tool_call.id,
            )
        )
    return True



# Copyright (c) 2025 Bytedance Ltd. and/or its affiliates
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
from contextlib import asynccontextmanager
from typing import Any, AsyncIterator, Dict, Generic, Optional, Tuple, Type, Union

import fastapi
import uvicorn
from fastapi import FastAPI, HTTPException
from fastapi.middleware.cors import CORSMiddleware
from pydantic import BaseModel, Field
from starlette.responses import StreamingResponse
from volcenginesdkarkruntime._exceptions import ArkAPIError

from arkitect.core.client import Client, get_client_pool, load_request
from arkitect.core.component.llm import ArkChatRequest
from arkitect.core.errors import APIException, ArkError, InternalServiceError
from arkitect.core.runtime import AsyncRunner, RequestType, ResponseType

from .middleware import (
    ListenDisconnectionMiddleware,
    LogIdMiddleware,
)


def _get_lock() -> asyncio.Lock:
    return asyncio.Lock()


def _default_endpoint_config() -> Dict[str, Type[ArkChatRequest]]:
    return {"/api/v3/bots/chat/completions": ArkChatRequest}


def _default_healthcheck_config() -> str:
    return "/healthz"


class BotServer(BaseModel, Generic[RequestType, ResponseType]):
    """BotServer in charge of the server router and runtime"""

    runner: AsyncRunner
    """ handler runtime of bot server"""
    endpoint_config: Dict[str, Type[RequestType]] = Field(  # type: ignore[arg-type]
        default_factory=_default_endpoint_config  # type: ignore[arg-type]
    )  # type: ignore[arg-type]
    """
    configuration of endpoint path & request class,
    format: {$ep_path: $request_cls }
    """
    health_check_path: str = Field(default_factory=_default_healthcheck_config)
    """path for server health check"""
    app: FastAPI
    """server application"""

    class Config:
        """Configuration for this pydantic object."""

        arbitrary_types_allowed = True

    def __init__(
        self,
        runner: AsyncRunner,
        endpoint_config: Optional[Dict[str, Type[RequestType]]] = None,
        clients: Optional[Dict[str, Tuple[Type[Client], Dict[str, Any]]]] = None,
        app: Optional[FastAPI] = None,
        health_check_path: Optional[str] = None,
        **kwargs: Any,
    ):
        @asynccontextmanager
        async def lifespan(app: FastAPI) -> AsyncIterator[Dict[str, Any]]:
            yield {"client_pool": get_client_pool(clients)}

        super().__init__(
            runner=runner,
            endpoint_config=endpoint_config or _default_endpoint_config(),
            health_check_path=health_check_path or _default_healthcheck_config(),
            app=app or FastAPI(lifespan=lifespan),
            **kwargs,
        )

        self.add_middlewares(self.app)
        self.add_routes(self.app)

    async def handler(self, http_request: fastapi.Request) -> fastapi.Response:
        try:
            request: RequestType = await load_request(
                http_request=http_request,
                req_cls=self.get_request_cls(http_request.url.path),
            )

            if request.stream:
                generator = self.runner.astream(request)
                return StreamingResponse(generator, media_type="text/event-stream")
            else:
                return await self.runner.arun(request)

        except APIException as e:
            raise HTTPException(
                status_code=e.http_code,
                detail=e.to_error().model_dump(exclude_unset=True, exclude_none=True),
            )
        except ArkAPIError as e:
            raise HTTPException(
                status_code=e.status_code if hasattr(e, "status_code") else 500,
                detail=ArkError(
                    code=e.code, message=e.message, param=e.param, type=e.type
                ).model_dump(exclude_unset=True, exclude_none=True),
            )
        except Exception as e:
            err = InternalServiceError(str(e))
            raise HTTPException(
                status_code=err.http_code,
                detail=err.to_error().model_dump(exclude_none=True, exclude_unset=True),
            )

    async def health_check(self) -> Any:
        return {}

    def add_routes(self, app: FastAPI) -> None:
        for endpoint_path, request_cls in self.endpoint_config.items():
            app.add_api_route(
                endpoint_path,
                self.handler,
                methods=["POST", "OPTIONS", "GET"],
            )
        app.add_api_route(
            self.health_check_path,
            self.health_check,
            methods=["GET"],
        )

    def get_request_cls(self, api_path: str) -> Type[RequestType]:
        assert api_path in self.endpoint_config, ValueError(
            "no corresponding api in endpoint configs"
        )
        return self.endpoint_config[api_path]

    @staticmethod
    def add_middlewares(app: FastAPI) -> None:
        app.add_middleware(ListenDisconnectionMiddleware)
        app.add_middleware(LogIdMiddleware)
        app.add_middleware(
            CORSMiddleware,
            allow_origins=["*"],
            allow_credentials=True,
            allow_methods=["*"],
            allow_headers=["*"],
        )

    @staticmethod
    def run(
        app: Union[FastAPI, str],
        host: str = "0.0.0.0",
        port: int = 8080,
        workers_num: int = 1,
        **kwargs: Any,
    ) -> None:
        uvicorn.run(app, host=host, port=port, workers=workers_num, **kwargs)



# Copyright (c) 2025 Bytedance Ltd. and/or its affiliates
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import inspect
from inspect import Signature
from typing import Any, AsyncIterable, Callable, Dict, Tuple, Type

from volcenginesdkarkruntime import AsyncArk

from arkitect.core.client import Client
from arkitect.core.component.llm import (
    ArkChatRequest,
)
from arkitect.core.runtime import (
    AsyncRunner,
    ChatAsyncRunner,
    CustomAsyncRunner,
    Request,
    RequestType,
    ResponseType,
)
from arkitect.telemetry.trace import task


@task()
def get_runner(
    runnable_func: Callable[[RequestType], AsyncIterable[ResponseType]],
) -> AsyncRunner:
    signature = inspect.signature(runnable_func)
    request_cls: Type[RequestType] = get_request_cls(signature)
    response_cls: Type[ResponseType] = get_response_cls(signature)

    if issubclass(request_cls, ArkChatRequest):
        return ChatAsyncRunner(runnable_func)  # type: ignore
    else:
        return CustomAsyncRunner(response_cls, runnable_func)


@task()
def get_endpoint_config(
    endpoint_path: str,
    runnable_func: Callable[[RequestType], AsyncIterable[ResponseType]],
) -> Dict[str, Type[RequestType]]:
    signature = inspect.signature(runnable_func)
    return {endpoint_path: get_request_cls(signature)}


@task()
def get_request_cls(signature: Signature) -> Type[RequestType]:
    parameters = signature.parameters.values()

    for param in parameters:
        annotation = param.annotation
        assert issubclass(annotation, Request), TypeError(
            "function request should be subclass of request"
        )
        return annotation

    raise Exception("should not reach here")


@task()
def get_response_cls(signature: Signature) -> Type[ResponseType]:
    return_cls = signature.return_annotation
    assert hasattr(return_cls, "__origin__") and issubclass(
        return_cls.__origin__, AsyncIterable
    ), TypeError("function response should be AsyncIterable")
    assert return_cls.__args__ and len(return_cls.__args__) == 1, TypeError(
        "function should return one value"
    )

    # skip response class check
    response_cls = return_cls.__args__[0]
    return response_cls


def get_default_client_configs() -> Dict[str, Tuple[Type[Client], Any]]:
    return {
        "chat": (
            AsyncArk,
            {
                "region": "cn-beijing",
                "ak": "xxx",
                "sk": "xxx",
                "api_key": "xxx",
            },
        ),
    }



# Copyright (c) 2025 Bytedance Ltd. and/or its affiliates
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, List, Optional

from .manifest import ToolManifest


def tool_key(action_name: str, tool_name: str) -> str:
    return action_name + "/" + tool_name


class ToolPool:
    _tools: Dict[str, ToolManifest] = {}

    @classmethod
    def register(cls, tm: ToolManifest) -> None:
        if not tm:
            return

        cls._tools[tool_key(tm.action_name, tm.tool_name)] = tm

    @classmethod
    def all(cls, tool_names: Optional[List[str]] = None) -> Dict[str, ToolManifest]:
        """ """
        if not tool_names:
            return cls._tools

        return {name: tm for name, tm in cls._tools.items() if name in tool_names}

    @classmethod
    def get(cls, action_name: str, tool_name: str) -> Optional[ToolManifest]:
        """ """
        return cls._tools.get(tool_key(action_name, tool_name))



# Copyright (c) 2025 Bytedance Ltd. and/or its affiliates
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Message version and header size
PROTOCAL_VERSION = 0b1
HEADER_SIZE = 0b1

# Message Type:
INVALID = 0b0000
FULL_CLIENT = 0b0001
AUDIO_ONLY_CLIENT = 0b0010
FULL_SERVER = 0b0011
AUDIO_ONLY_SERVER = 0b0100
FRONT_END_RESULT_SERVER = 0b0101
ERROR = 0b0110
SERVER_ACK = AUDIO_ONLY_SERVER

# Message Type Specific Flags
NO_SEQUENCE = 0b0000  # no check sequence
POS_SEQUENCE = 0b0001
LAST_NO_SEQUENCE = 0b0010
NEG_SEQUENCE = 0b0011
WITH_EVENT = 0b0100


# Message Serialization
NO_SERIALIZATION = 0b0000
JSON = 0b0001

# Message Compression
NO_COMPRESSION = 0b0000
GZIP = 0b0001


# 默认事件,对于使用事件的方案，可以通过非0值来校验事件的合法性
EventNone = 0


EventStartConnection = 1
EventFinishConnection = 2


EventConnectionStarted = 50
EventConnectionFailed = 51
EventConnectionFinished = 52


EventStartSession = 100
EventFinishSession = 102

EventSessionStarted = 150
EventSessionFinished = 152
EventSessionFailed = 153

EventTaskRequest = 200

EventTTSSentenceStart = 350
EventTTSSentenceEnd = 351
EventTTSResponse = 352


NAMESPACE = "BidirectionalTTS"
INT_SIZE = 4
DEFAULT_SPEAKER = "zh_female_tianmeixiaoyuan_moon_bigtts"



# Copyright (c) 2025 Bytedance Ltd. and/or its affiliates
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict, List, Optional, Union

from langchain.prompts.chat import BaseChatPromptTemplate
from langchain.schema.output_parser import BaseTransformOutputParser
from volcenginesdkarkruntime import Ark, AsyncArk
from volcenginesdkarkruntime._streaming import AsyncStream
from volcenginesdkarkruntime.types.chat import (
    ChatCompletion,
    ChatCompletionChunk,
)

from arkitect.core.component.tool import ToolManifest
from arkitect.telemetry.trace import task
from arkitect.utils.context import get_extra_headers

from .base import BaseLanguageModel
from .function_call import handle_function_call
from .model import (
    ArkChatCompletionChunk,
    ArkChatParameters,
    ArkChatRequest,
    ArkChatResponse,
    ArkMessage,
    FunctionCallMode,
)
from .utils import format_ark_prompts


class BaseChatLanguageModel(BaseLanguageModel):
    messages: List[ArkMessage]
    parameters: Optional[ArkChatParameters] = None
    template: Optional[BaseChatPromptTemplate] = None
    output_parser: Optional[BaseTransformOutputParser] = None

    class Config:
        """Configuration for this pydantic object."""

        arbitrary_types_allowed = True

    def generate_prompts(
        self,
        messages: List[ArkMessage],
        *,
        additional_system_prompts: Optional[List[str]] = None,
        **kwargs: Any,
    ) -> List[ArkMessage]:
        # additional system prompts would be put at first
        if additional_system_prompts:
            prompts = [
                ArkMessage(role="system", content=system_prompt)
                for system_prompt in additional_system_prompts
            ]
            prompts.extend(messages)
            messages = prompts

        if not self.template:
            return messages

        return format_ark_prompts(self.template, messages, **kwargs)

    def get_request_model(self, **kwargs: Any) -> str:
        return self.endpoint_id

    @task()
    def parse_output(self, text: str) -> Any:
        if not self.output_parser:
            return text

        return self.output_parser.parse(text)

    @task()
    async def aparse_output(self, text: str) -> Any:
        if not self.output_parser:
            return text

        return await self.output_parser.aparse(text)

    @task()
    async def _arun(
        self,
        request: ArkChatRequest,
        extra_headers: Optional[Dict[str, str]] = {},
        extra_query: Optional[Dict[str, Any]] = None,
        extra_body: Optional[Dict[str, Any]] = None,
    ) -> Union[ChatCompletion, AsyncStream[ChatCompletionChunk]]:
        assert isinstance(self.client, AsyncArk), TypeError("Invalid Client for v3 sdk")

        params = request.get_chat_request(extra_body)

        extra_headers = get_extra_headers(extra_headers)

        return await self.client.chat.completions.create(
            **params,
            extra_headers=extra_headers,
            extra_query=extra_query,
        )

    @task()
    def _run(
        self,
        request: ArkChatRequest,
        extra_headers: Optional[Dict[str, str]] = {},
        extra_query: Optional[Dict[str, Any]] = None,
        extra_body: Optional[Dict[str, Any]] = None,
    ) -> Union[ChatCompletion, AsyncStream[ChatCompletionChunk]]:
        sync_client = Ark()

        extra_headers = get_extra_headers(extra_headers)

        return sync_client.chat.completions.create(
            **request.get_chat_request(extra_body),
            extra_headers=extra_headers,
            extra_query=extra_query,
        )

    def run(
        self,
        extra_headers: Optional[Dict[str, str]] = {},
        extra_query: Optional[Dict[str, Any]] = None,
        extra_body: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ) -> Any:
        """
        Runs a chat request and returns the response.
        """
        parameters = (
            self.parameters.model_dump(exclude_none=True, exclude_unset=True)
            if self.parameters
            else {}
        )
        request = ArkChatRequest(
            stream=False,
            messages=self.generate_prompts(self.messages, **kwargs),
            model=self.get_request_model(**kwargs),
            **parameters,
        )

        completion = self._run(request, extra_headers, extra_query, extra_body)
        return ArkChatResponse(**completion.__dict__)

    def stream(
        self,
        extra_headers: Optional[Dict[str, str]] = {},
        extra_query: Optional[Dict[str, Any]] = None,
        extra_body: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ) -> Any:
        """
        Streams chat completions from the language model.
        """
        parameters = (
            self.parameters.model_dump(exclude_none=True, exclude_unset=True)
            if self.parameters
            else {}
        )
        request = ArkChatRequest(
            stream=True,
            messages=self.generate_prompts(self.messages, **kwargs),
            model=self.get_request_model(**kwargs),
            **parameters,
        )

        completion = self._run(request, extra_headers, extra_query, extra_body)
        for resp in completion:
            yield ArkChatCompletionChunk(**resp.__dict__)

    async def arun(
        self,
        extra_headers: Optional[Dict[str, str]] = {},
        extra_query: Optional[Dict[str, Any]] = None,
        extra_body: Optional[Dict[str, Any]] = None,
        *,
        functions: Optional[Dict[str, ToolManifest]] = None,
        function_call_mode: Optional[FunctionCallMode] = FunctionCallMode.SEQUENTIAL,
        additional_system_prompts: Optional[List[str]] = None,
        **kwargs: Any,
    ) -> ArkChatResponse:
        """
        Asynchronously runs a chat request and returns the response.
        """
        parameters = (
            self.parameters.model_dump(exclude_none=True, exclude_unset=True)
            if self.parameters
            else {}
        )
        request = ArkChatRequest(
            stream=False,
            messages=self.generate_prompts(
                self.messages,
                additional_system_prompts=additional_system_prompts,
                **kwargs,
            ),
            model=self.get_request_model(**kwargs),
            **parameters,
        )
        responses = []
        while True:
            completion: ChatCompletion = await self._arun(
                request, extra_headers, extra_query, extra_body
            )
            responses.append(completion)

            if completion.choices and completion.choices[0].finish_reason:
                if not await handle_function_call(
                    request, completion, functions, function_call_mode
                ):
                    break

        return ArkChatResponse.merge(responses)

    async def astream(
        self,
        extra_headers: Optional[Dict[str, str]] = {},
        extra_query: Optional[Dict[str, Any]] = None,
        extra_body: Optional[Dict[str, Any]] = None,
        *,
        functions: Optional[Dict[str, ToolManifest]] = None,
        function_call_mode: Optional[FunctionCallMode] = FunctionCallMode.SEQUENTIAL,
        additional_system_prompts: Optional[List[str]] = None,
        **kwargs: Any,
    ) -> AsyncStream[ArkChatCompletionChunk]:
        """
        Asynchronously streams chat completions from the language model.
        """
        parameters = (
            self.parameters.model_dump(exclude_none=True, exclude_unset=True)
            if self.parameters
            else {}
        )
        request = ArkChatRequest(
            stream=True,
            messages=self.generate_prompts(
                self.messages,
                additional_system_prompts=additional_system_prompts,
                **kwargs,
            ),
            model=self.get_request_model(**kwargs),
            **parameters,
        )

        usage_chunks, cumulated = [], []
        while True:
            completion = await self._arun(
                request, extra_headers, extra_query, extra_body
            )
            # default: one iter
            is_more_request = False
            async for resp in completion:  # type: ChatCompletionChunk
                if resp.usage:
                    usage_chunks.append(resp)
                else:
                    # cumulated chunks is used for caculator/fc inner cot output
                    cumulated.append(resp)
                    yield ArkChatCompletionChunk(**resp.__dict__)

                if resp.choices and resp.choices[0].finish_reason:
                    ark_resp = ArkChatCompletionChunk(**resp.__dict__).merge(cumulated)
                    # clear after used
                    cumulated = []

                    is_more_request = await handle_function_call(
                        request, ark_resp, functions, function_call_mode
                    )
                    if (
                        request.stream_options
                        and request.stream_options.get("include_usage") is True
                    ):
                        usage_chunks.append(ark_resp)

            if not is_more_request:
                break

        if len(usage_chunks) > 0:
            yield ArkChatCompletionChunk.merge(usage_chunks)



# Copyright (c) 2025 Bytedance Ltd. and/or its affiliates
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
from typing import (
    Any,
    AsyncIterable,
    AsyncIterator,
    Coroutine,
    List,
    Optional,
    Tuple,
    TypeVar,
)

T = TypeVar("T")


async def aenumerate(
    asequence: AsyncIterable[T], start: int = 0
) -> AsyncIterator[Tuple[int, T]]:
    """Asynchronously enumerate an async iterator from a given start value"""
    n = start
    async for elem in asequence:
        yield n, elem
        n += 1


async def anext(asequence: AsyncIterator[T]) -> T:
    """
    Asynchronously get next item from async generator
    Note: python3.10 has a anext builtin function
    """
    return await asequence.__anext__()


async def gather(*coros_or_futures: Coroutine) -> List[Any]:
    tasks: List[asyncio.Task[Any]] = [
        asyncio.create_task(coro=task) for task in coros_or_futures
    ]
    try:
        results = await asyncio.gather(*tasks)
    except Exception:
        for task in tasks:
            task.cancel()
        raise
    else:
        return results


class AsyncTimedIterable:
    def __init__(
        self,
        iterable: AsyncIterable[Any],
        timeout: Optional[int] = None,
        sentinel: Optional[Any] = None,
    ):
        class AsyncTimedIterator:
            def __init__(self) -> None:
                self._iterator = iterable.__aiter__()

            async def __anext__(self) -> Any:
                try:
                    return await asyncio.wait_for(self._iterator.__anext__(), timeout)
                except asyncio.TimeoutError as e:
                    if sentinel:
                        raise sentinel
                    else:
                        raise e

        self._factory = AsyncTimedIterator

    def __aiter__(self) -> Any:
        return self._factory()



# Copyright (c) 2025 Bytedance Ltd. and/or its affiliates
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import uuid
from typing import Any, AsyncIterable, Optional, Union

import websockets

from arkitect.core.component.llm.model import ArkChatCompletionChunk, ArkChatResponse
from arkitect.core.component.tts.base import AsyncBaseTTSClient, TTSResponseChunk
from arkitect.core.component.tts.constants import (
    NAMESPACE,
    EventFinishSession,
    EventStartConnection,
    EventStartSession,
    EventTaskRequest,
    EventTTSSentenceStart,
)
from arkitect.core.component.tts.model import (
    ConnectionParams,
    Message,
    ResponseEvent,
    TextRequest,
    TTSRequest,
)
from arkitect.core.component.tts.utils import parse_response
from arkitect.core.errors import InvalidParameter
from arkitect.telemetry.logger import ERROR, INFO
from arkitect.telemetry.trace import task

__all__ = ["AsyncTTSClient"]


class AsyncTTSClient(AsyncBaseTTSClient):
    def __init__(
        self,
        access_key: str,
        app_key: str,
        connection_params: ConnectionParams,
        api_resource_id: str = "volc.service_type.10029",
        conn_id: str = str(uuid.uuid4()),
        log_id: str = str(uuid.uuid4()),
        base_url: str = "wss://openspeech.bytedance.com/api/v3/tts/bidirection",
    ):
        self.api_resource_id = api_resource_id
        self.access_key = access_key
        self.app_key = app_key
        self.conn_id = conn_id
        self.log_id = log_id
        self.base_url = base_url

        self.conn: Optional[websockets.WebSocketClientProtocol] = None
        self.session_id: Optional[str] = None
        self.connection_params: ConnectionParams = connection_params
        self.inited = False

    async def init(
        self,
        namespace: str = NAMESPACE,
    ) -> None:
        headers = self._build_http_header()
        INFO("with logID: %s , header: %s", self.log_id, headers)
        self.conn = await websockets.connect(self.base_url, extra_headers=headers)
        INFO("Dial server with LogID: %s", self.log_id)
        # Create a new message with type MsgTypeFullClient and flag MsgTypeFlagWithEvent
        msg = Message(event=EventStartConnection)
        frame = msg.write_start_connection()
        await self._send_frame(frame)
        # Read ConnectionStarted message
        response = await self.conn.recv()
        parse_response(response)
        await self._start_tts_session(
            namespace=namespace,
            params=self.connection_params,
        )
        self.inited = True

    async def reset_conn(self) -> None:
        await self.close()
        await self.init()
        INFO("TTS Reset connection")

    async def _start_tts_session(
        self, namespace: str, params: ConnectionParams
    ) -> ResponseEvent:
        msg = Message(
            connection_id=self.conn_id,
            event=EventStartSession,
        )
        # Marshal the message into a binary frame
        p = params.model_dump(mode="json")
        session_config = {
            "event": EventStartSession,
            "namespace": namespace,
            "req_params": p,
        }
        msg.payload = session_config
        frame = msg.write_start_tts_session()
        await self._send_frame(frame)
        if self.conn is None:
            raise ValueError("TTS connection is not established")
        response = await self.conn.recv()
        result = parse_response(response)
        self.session_id = result.session_id
        return result

    async def _send_text_data(self, data: TextRequest) -> None:
        req_params = self.connection_params.model_dump(mode="json")
        req_params["text"] = data.text
        req = TTSRequest(
            event=EventTaskRequest,
            req_params=req_params,
        )
        msg = Message(
            event=EventTaskRequest,
            session_id=self.session_id,
        )
        msg.payload = req.model_dump(mode="json")
        # Marshal the message into a binary frame
        frame = msg.write_text_request()
        await self._send_frame(frame)
        if data.finished:
            await self._send_finish_session()

    async def _send_frame(self, frame: bytes) -> None:
        if self.conn is None:
            raise ValueError("Connection is not established")
        await self.conn.send(frame)

    async def _send_finish_session(self) -> None:
        msg = Message(
            event=EventFinishSession,
            session_id=self.session_id,
        )
        msg.payload = {}
        frame = msg.write_finish_session()
        await self._send_frame(frame)

    async def _receive_data(self) -> ResponseEvent:
        if self.conn is None:
            raise ValueError("Connection is not established")
        response = await self.conn.recv()
        result = parse_response(response)
        return result

    async def _receive_audio_data(self) -> ResponseEvent:
        if self.conn is None:
            raise ValueError("Connection is not established")
        # Implement the receive audio data logic here
        while True:
            response = await self.conn.recv()
            result = parse_response(response)
            if result.audio_only or result.session_finished:
                return result

    async def close(self) -> None:
        if self.conn is not None:
            await self.conn.close()
            self.conn = None
        self.inited = False

    def _build_http_header(self) -> dict:
        headers = {
            "X-Tt-Logid": self.log_id,
            "X-Api-Resource-Id": self.api_resource_id,
            "X-Api-Access-Key": self.access_key,
            "X-Api-App-Key": self.app_key,
            "X-Api-Connect-Id": self.conn_id,
        }
        return headers

    async def _get_tts_stream(
        self,
        include_transcript: bool,
    ) -> AsyncIterable[TTSResponseChunk]:
        """

        A tts stream consist of interleaving audio and transcript.
        The sequence is as follow
        <TRANSCRIPT 1> <AUDIO 1> <AUDIO 1> ... <TRANSCRIPT 1>
        <TRANSCRIPT 2> <AUDIO 2> <AUDIO 2> ...<TRANSCRIPT 2>

        Each sentence of transcript will be returned twice,
        indicating the start and end of its corresponding andio chunks.
        We remove the second transcript chunk to avoid duplicate
        """
        while True:
            response = await self._receive_data()
            if response.audio_only:
                yield TTSResponseChunk(event=response.event, audio=response.audio)
            if response.session_finished:
                yield TTSResponseChunk(event=response.event)
                break
            if response.event == EventTTSSentenceStart and include_transcript:
                transcript = response.payload_msg.get("text", "")
                yield TTSResponseChunk(event=response.event, transcript=transcript)
            else:
                yield TTSResponseChunk(event=response.event)

    async def tts(
        self,
        source: Union[
            AsyncIterable[Union[ArkChatCompletionChunk, ArkChatResponse, str]], str
        ],
        stream: bool = True,
        **kwargs: Any,
    ) -> AsyncIterable[TTSResponseChunk]:
        try:
            if not self.inited:
                await self.init()
            include_transcript: bool = kwargs.get("include_transcript", True)

            @task()
            async def send_text_to_tts(
                text_stream: Union[
                    AsyncIterable[Union[ArkChatCompletionChunk, ArkChatResponse, str]],
                    str,
                ],
            ) -> None:
                if isinstance(source, str):
                    await self._send_text_data(
                        TextRequest(
                            text=source,
                            finished=True,
                        )
                    )
                else:
                    async for resp in text_stream:
                        if isinstance(resp, ArkChatCompletionChunk):
                            text = resp.choices[0].delta.content
                        elif isinstance(resp, ArkChatResponse):
                            text = resp.choices[0].message.content
                        elif isinstance(resp, str):
                            text = resp
                        else:
                            raise InvalidParameter(f"Invalid type: {type(resp)}")
                        await self._send_text_data(
                            TextRequest(
                                text=text,
                                finished=False,
                            )
                        )
                    await self._send_finish_session()

            t = asyncio.create_task(send_text_to_tts(text_stream=source))
            audio_part: bytes = b""
            audio_transcript: str = ""
            async for chunk in self._get_tts_stream(include_transcript):
                if stream:
                    yield chunk
                if chunk.audio:
                    audio_part += chunk.audio
                if chunk.transcript:
                    audio_transcript += chunk.transcript
            if not stream:
                yield TTSResponseChunk(audio=audio_part, transcript=audio_transcript)
            await t
            await self.close()
        except Exception as e:
            ERROR(str(e))
        finally:
            await self.close()



# Copyright (c) 2025 Bytedance Ltd. and/or its affiliates
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc
import json
import logging
from typing import (
    Any,
    AsyncIterable,
    AsyncIterator,
    Callable,
    Coroutine,
    Generic,
    Type,
    Union,
)

from pydantic import BaseModel, ValidationError
from volcenginesdkarkruntime._exceptions import ArkAPIError

from arkitect.core.errors import (
    APIException,
    ArkError,
    InternalServiceError,
    parse_pydantic_error,
)

from .model import RequestType, Response, ResponseType


class AsyncRunner(BaseModel, Generic[RequestType, ResponseType]):
    invoke: Callable[[RequestType], Coroutine[Any, Any, AsyncIterable[ResponseType]]]

    class Config:
        """Configuration for this pydantic object."""

        arbitrary_types_allowed = True

    def __init__(
        self,
        runnable_func: Callable[[RequestType], AsyncIterable[ResponseType]],
        **kwargs: Any,
    ):
        async def handler(
            request: RequestType,
        ) -> AsyncIterable[ResponseType]:
            return runnable_func(request)

        super().__init__(invoke=handler, **kwargs)

    @abc.abstractmethod
    async def arun(self, request: RequestType) -> ResponseType:
        pass

    @abc.abstractmethod
    def astream(self, request: RequestType) -> AsyncIterator[str]:
        pass


class CustomAsyncRunner(AsyncRunner[RequestType, ResponseType]):
    response_cls: Type[ResponseType]

    def __init__(
        self,
        response_cls: Type[ResponseType],
        runnable_func: Callable[[RequestType], AsyncIterable[ResponseType]],
        **kwargs: Any,
    ):
        super().__init__(
            response_cls=response_cls, runnable_func=runnable_func, **kwargs
        )

    async def arun(self, request: RequestType) -> ResponseType:  # type: ignore
        try:
            async for resp in await self.invoke(request):  # type: ResponseType
                return resp
        except APIException as e:
            resp = self.response_cls(error=e.to_error())
            logging.error(f"bot meet error{e}")
            return resp
        except Exception as e:
            err = InternalServiceError(str(e))
            resp = self.response_cls(error=err.to_error())
            logging.error(f"bot meet internal error{err}")
            return resp

    async def astream(self, request: RequestType) -> AsyncIterator[str]:  # type: ignore
        try:
            async for resp in await self.invoke(request):  # type: ResponseType
                if isinstance(resp, BaseModel):
                    yield f"data:{resp.model_dump_json(exclude_unset=True, exclude_none=True)}\r\n\r\n"  # noqa E501
                else:
                    yield f"data:{json.dumps(resp, ensure_ascii=False)}\r\n\r\n"  # noqa E501
        except APIException as e:
            resp = self.response_cls(error=e.to_error())
            logging.error("stream chat meet error")
            yield f"data:{resp.model_dump_json(exclude_unset=True, exclude_none=True)}\r\n\r\n"  # noqa E501
        except Exception as e:
            err = InternalServiceError(str(e))
            resp = self.response_cls(error=err.to_error())
            logging.error("stream chat meet error")
            yield f"data:{resp.model_dump_json(exclude_unset=True, exclude_none=True)}\r\n\r\n"  # noqa E501
        yield "data:[DONE]\r\n\r\n"


class ChatAsyncRunner(AsyncRunner[RequestType, ResponseType]):
    async def arun(self, request: RequestType) -> Union[Response, ResponseType]:  # type: ignore
        try:
            async for resp in await self.invoke(request):  # type: ResponseType
                return resp
        except APIException as e:
            logging.error(f"[API Error]: chat meet error:{e}")
            raise e
        except ValidationError as e:
            logging.error(f"[Validation Error]: chat meet parameter error:{e}")
            raise parse_pydantic_error(e)
        except ArkAPIError as e:
            logging.error(f"[Calling Chat Error]: chat meet error:{e}")
            raise e
        except Exception as e:
            err = InternalServiceError(str(e))
            logging.error(f"[Internal Error]: chat meet error:{e}")
            raise err

    async def astream(self, request: RequestType) -> AsyncIterator[str]:
        try:
            async for resp in await self.invoke(request):  # type: ResponseType
                yield f"data:{resp.model_dump_json(exclude_none=True)}\r\n\r\n"  # noqa E501
        except APIException as e:
            err = Response(error=e.to_error())
            logging.error(f"[API Error]: stream chat meet error:{e}")
            yield f"data:{err.model_dump_json(exclude_unset=True, exclude_none=True)}\r\n\r\n"  # noqa E501
        except ValidationError as e:
            err = Response(error=parse_pydantic_error(e).to_error())
            logging.error(f"[Validation Error]: stream chat meet parameter error:{e}")
            yield f"data:{err.model_dump_json(exclude_unset=True, exclude_none=True)}\r\n\r\n"  # noqa E501
        except ArkAPIError as e:
            err = Response(
                error=ArkError(
                    code=e.code or "",  # code is none when calling top failed
                    message=e.message or "",
                    type=e.type,
                    param=e.param,
                )
            )
            logging.error(f"[Calling Chat Error]: stream chat meet error:{e}")
            yield f"data:{err.model_dump_json(exclude_unset=True, exclude_none=True)}\r\n\r\n"  # noqa E501
        except Exception as e:
            err = Response(error=InternalServiceError(str(e)).to_error())
            logging.error(f"[Internal Error]: stream chat meet error:{e}")
            yield f"data:{err.model_dump_json(exclude_unset=True, exclude_none=True)}\r\n\r\n"  # noqa E501
        yield "data:[DONE]\r\n\r\n"



# Copyright (c) 2025 Bytedance Ltd. and/or its affiliates
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import json
import unittest
from typing import Any, AsyncIterator, Callable, Dict, Tuple, Type
from unittest.mock import MagicMock, patch

from arkitect.core.client import Client
from arkitect.core.errors import InvalidParameter
from arkitect.core.runtime import ChatAsyncRunner, Context, Request, Response
from arkitect.launcher.vefaas.wrapper import (
    APIException,
    Environment,
    MissingParameter,
    _get_parameters,
    bot_wrapper,
    parse_function_request,
    parse_function_response,
)


class TestBotWrapper(unittest.TestCase):
    @staticmethod
    def mock_client() -> Tuple[Type[Client], Any]:
        return MagicMock(spec=Client), MagicMock()

    @staticmethod
    def mock_event_context() -> Tuple[Dict[str, Any], Context]:
        event = {
            "body": json.dumps(
                {
                    "stream": False,
                    "messages": [{"role": "user", "content": "查今天的新闻"}],
                    "model": "bot-123",
                    "stream_options": {"include_usage": True},
                }
            )
        }
        context = Context(request_id="123")
        return event, context

    @staticmethod
    def mock_runner(func: Callable) -> ChatAsyncRunner:
        return ChatAsyncRunner(runnable_func=func)  # type: ignore

    @staticmethod
    def mock_request(endpoint_config: Dict[str, Any], event: Any) -> Request:
        request = Request(stream=False)
        return request

    @staticmethod
    def mock_parse_request(event: Any, request_cls: Type[Request]) -> Request:
        return Request(stream=False)

    @staticmethod
    def mock_parse_response(status_code: int, content: Any) -> Dict[str, Any]:
        return {"key": "value"}

    @staticmethod
    async def handler(request: Request) -> AsyncIterator[Response]:
        yield Response()

    def test_bot_wrapper_entry(self) -> None:
        clients = {"client": self.mock_client()}
        with (
            patch("ark.launcher.vefaas.wrapper.initialize") as mock_initialize,
            patch(
                "ark.launcher.vefaas.wrapper.get_runner", side_effect=self.mock_runner
            ) as mock_get_runner,
            patch(
                "ark.launcher.vefaas.wrapper.get_endpoint_config"
            ) as mock_get_endpoint_config,
            patch(
                "ark.launcher.vefaas.wrapper.parse_request",
                side_effect=self.mock_parse_request,
            ) as mock_parse_request,
            patch(
                "ark.launcher.vefaas.wrapper.parse_response",
                side_effect=self.mock_parse_response,
            ) as mock_parse_response,
        ):
            wrapped_func = bot_wrapper(
                endpoint_path="/api/v3/bots/chat/completions",
                clients=clients,
                trace_on=False,
            )(self.handler)
            event, context = self.mock_event_context()
            result = asyncio.run(wrapped_func(event, context))

            mock_initialize.assert_called_once_with(context, clients, False, None)
            mock_get_runner.assert_called_once_with(self.handler)
            mock_get_endpoint_config.assert_called_once_with(
                endpoint_path="/api/v3/bots/chat/completions",
                runnable_func=self.handler,
            )
            mock_parse_request.assert_called_once_with(
                event=event,
                request_cls=mock_get_endpoint_config.return_value.get.return_value,
            )
            mock_parse_response.assert_called_once_with(status_code=200, content="{}")
            self.assertIsInstance(result, dict)
            self.assertEqual(result, {"key": "value"})

    def test_get_parameters_local(self) -> None:
        # Test case for LOCAL environment
        request_data = {"key": "value"}
        result = _get_parameters(Environment.LOCAL, request_data)
        self.assertEqual(result, (request_data, Context()))

        with self.assertRaises(MissingParameter):
            _get_parameters(
                Environment.LOCAL
            )  # Missing parameters should raise MissingParameter

    def test_get_parameters_vefaas(self) -> None:
        # Test case for VEFAAS environment
        event_data = {"body": "value"}
        context_data = Context()
        result = _get_parameters(Environment.VEFAAS, event_data, context_data)
        self.assertEqual(result, (event_data, context_data))

        with self.assertRaises(TypeError):
            _get_parameters(
                Environment.VEFAAS
            )  # Missing parameters should raise TypeError

    def test_parse_function_request_local_dict(self) -> None:
        environment = Environment.LOCAL
        parameters = {"key": "value"}
        endpoint_path = "/test"
        func = self.handler

        result = parse_function_request(environment, parameters, endpoint_path, func)
        self.assertIsInstance(result, Request)

    def test_parse_function_request_local_request_instance(self) -> None:
        environment = Environment.LOCAL
        request_instance = Request()
        endpoint_path = "/test"
        func = self.handler

        result = parse_function_request(
            environment, request_instance, endpoint_path, func
        )
        self.assertEqual(result, request_instance)

    def test_parse_function_request_vefaas(self) -> None:
        environment = Environment.VEFAAS
        parameters = {"body": json.dumps({"stream": False})}
        endpoint_path = "/test"
        func = self.handler

        result = parse_function_request(environment, parameters, endpoint_path, func)
        self.assertIsInstance(result, Request)

    def test_parse_function_response_local_response(self) -> None:
        environment = Environment.LOCAL
        response = Response()

        result = parse_function_response(environment, response=response)
        self.assertEqual(result, response)

    def test_parse_function_response_local_exception(self) -> None:
        environment = Environment.LOCAL
        exception = InvalidParameter(parameter="test")

        with self.assertRaises(APIException):
            parse_function_response(environment, exception=exception)

    def test_parse_function_response_vefaas_response(self) -> None:
        environment = Environment.VEFAAS
        response = Response()

        result = parse_function_response(environment, response=response)
        self.assertIsInstance(result, dict)
        self.assertIn("statusCode", result)
        self.assertIn("body", result)

    def test_parse_function_response_vefaas_exception(self) -> None:
        environment = Environment.VEFAAS
        exception = InvalidParameter(parameter="test")

        result = parse_function_response(environment, exception=exception)
        self.assertIsInstance(result, dict)
        self.assertIn("statusCode", result)
        self.assertIn("body", result)


if __name__ == "__main__":
    unittest.main()



# Copyright (c) 2025 Bytedance Ltd. and/or its affiliates
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Type

import fastapi
from httpx import Timeout
from pydantic import ValidationError
from volcenginesdkarkruntime import AsyncArk

from arkitect.core.errors import InvalidParameter, parse_pydantic_error
from arkitect.core.runtime import RequestType

from .base import get_client_pool


def default_ark_client() -> AsyncArk:
    """
    Retrieves or creates an instance of the AsyncArk client.

    This function attempts to fetch an existing client from the client pool.
    If no client is found, it creates a new instance of the AsyncArk client
    with a specified timeout configuration.

    Returns:
        AsyncArk: An instance of the AsyncArk client.
    """
    client_pool = get_client_pool()
    client: AsyncArk = client_pool.get_client("ark")  # type: ignore
    if not client:
        client = AsyncArk(timeout=Timeout(connect=1.0, timeout=60.0))
    return client


async def load_request(
    http_request: fastapi.Request,
    req_cls: Type[RequestType],
) -> RequestType:
    """
    Loads and validates a request from a FastAPI HTTP request.
    """
    if "content-type" not in http_request.headers:
        raise InvalidParameter("Invalid request: missing content-type")
    content_type, body = (
        http_request.headers.get("content-type", ""),
        await http_request.body(),
    )
    media_type = content_type.split(";")[0].strip()
    try:
        if media_type == "application/json":
            return req_cls.model_validate_json(body)
    except ValidationError as e:
        raise parse_pydantic_error(e)
    raise InvalidParameter(f"Invalid request: invalid content-type={content_type}")



# Copyright (c) 2025 Bytedance Ltd. and/or its affiliates
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .asyncio import AsyncRunner, ChatAsyncRunner, CustomAsyncRunner
from .model import Context, Request, RequestType, Response, ResponseType
from .runner import load_function
from .sync import SyncRunner

__all__ = [
    "AsyncRunner",
    "CustomAsyncRunner",
    "ChatAsyncRunner",
    "SyncRunner",
    "Request",
    "Response",
    "RequestType",
    "ResponseType",
    "load_function",
    "Context",
]



# Copyright (c) 2025 Bytedance Ltd. and/or its affiliates
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"errors"

from enum import Enum
from typing import Optional

from pydantic import BaseModel

__all__ = ["ErrorCode"]


class ReprEnum(Enum):
    """
    Only changes the repr(), leaving str() and format() to the mixed-in type.
    """


class StrEnum(str, ReprEnum):
    """
    Enum where members are also (and must be) strings
    """

    def __new__(cls, *values: str) -> "StrEnum":
        "values must already be of type `str`"
        if len(values) > 3:
            raise TypeError("too many arguments for str(): %r" % (values,))
        if len(values) == 1:
            # it must be a string
            if not isinstance(values[0], str):
                raise TypeError("%r is not a string" % (values[0],))
        if len(values) >= 2:
            # check that encoding argument is a string
            if not isinstance(values[1], str):
                raise TypeError("encoding must be a string, not %r" % (values[1],))
        if len(values) == 3:
            # check that errors argument is a string
            if not isinstance(values[2], str):
                raise TypeError("errors must be a string, not %r" % (values[2]))
        value = str(*values)
        member = str.__new__(cls, value)
        member._value_ = value
        return member

    def _generate_next_value_(self, start, count, last_values) -> str:  # type: ignore
        """
        Return the lower-cased version of the member name.
        """
        return self.lower()


class ErrorCode(StrEnum):
    http_code: int
    message: str
    error_type: str

    def __new__(
        cls,
        code: str,
        http_code: int,
        message: Optional[str] = None,
        error_type: Optional[str] = None,
    ) -> "ErrorCode":
        error_code = str.__new__(cls, code)
        error_code._value_ = code
        error_code.http_code = http_code  # type: ignore
        error_code.message = message or ""  # type: ignore
        error_code.error_type = error_type or ""  # type: ignore
        return error_code

    MissingParameter = (
        "MissingParameter",
        400,
        "The request failed because it is missing required parameters",
        "BadRequest",
    )
    InvalidParameter = (
        "InvalidParameter",
        400,
        "A parameter specified in the request is not valid: {parameter}",
        "BadRequest",
    )
    ResourceNotFound = (
        "ResourceNotFound",
        404,
        "The specified resource is not found",
        "NotFound",
    )
    RateLimitExceeded = (
        "RateLimitExceeded",
        429,
        "The Requests Per Minute(RPM) limit of the associated {resource_type} for your "
        "account has been exceeded.",
        "TooManyRequests",
    )
    ServerOverloaded = (
        "ServerOverloaded",
        429,
        "The service: {service} is currently unable to handle "
        "additional requests due to server overload.",
        "TooManyRequests",
    )

    SensitiveContentDetected = (
        "SensitiveContentDetected",
        400,
        "The request failed because the input text may contain sensitive information.",
        "BadRequest",
    )

    AuthenticationError = (
        "AuthenticationError",
        401,
        "The API key in the request is missing or invalid.",
        "Unauthorized",
    )

    AccessDenied = (
        "AccessDenied",
        403,
        "The request failed because you do not have access to the requested resource.",
        "Forbidden",
    )

    AccountOverdueError = (
        "AccountOverdueError",
        403,
        "The request failed because your account has an overdue balance.",
        "Forbidden",
    )
    QuotaExceeded = (
        "QuotaExceeded",
        429,
        "Your account {account_id} has exhausted its free trial "
        "quota for {resource_type}.",
        "TooManyRequests",
    )

    InternalServiceError = (
        "InternalServiceError",
        500,
        "The service encountered an unexpected internal error.",
        "InternalServerError",
    )

    Unknown = ("Unknown", 500, "Unknown error")

    ModelLoadingError = (
        "ModelLoadingError",
        429,
        "The request cannot be processed at this time because "
        "the model is currently being loaded",
        "TooManyRequests",
    )

    APITimeoutError = (
        "APITimeoutError",
        500,
        "Request timed out",
        "InternalServerError",
    )


class Error(BaseModel):
    code: str
    code_n: Optional[int] = None
    message: str


class ArkError(BaseModel):
    code: str
    message: str
    param: Optional[str] = None
    type: Optional[str] = None



# Copyright (c) 2025 Bytedance Ltd. and/or its affiliates
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod
from typing import Any, AsyncIterable, Optional, Union

from pydantic import BaseModel

from arkitect.core.component.llm.model import ArkChatCompletionChunk, ArkChatResponse


class TTSResponseChunk(BaseModel):
    event: Optional[int] = None
    audio: Optional[bytes] = None
    transcript: Optional[str] = None


class AsyncBaseTTSClient(ABC):
    @abstractmethod
    async def tts(
        self,
        source: AsyncIterable[Union[ArkChatCompletionChunk, ArkChatResponse, str]],
        stream: bool = True,
        **kwargs: Any,
    ) -> AsyncIterable[TTSResponseChunk]:
        ...



# Copyright (c) 2025 Bytedance Ltd. and/or its affiliates
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .asyncio import AsyncTimedIterable, aenumerate, anext, gather
from .json import dump_json_str, dump_json_str_truncate, dump_json_truncate
from .merge import dict_merge, list_item_merge

__all__ = [
    "AsyncTimedIterable",
    "gather",
    "anext",
    "aenumerate",
    "dict_merge",
    "list_item_merge",
    "dump_json_str",
    "dump_json_str_truncate",
    "dump_json_truncate",
]



# Copyright (c) 2025 Bytedance Ltd. and/or its affiliates
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from typing import Any, Dict, Optional, Tuple, Type

from arkitect.core.client import Client, get_client_pool
from arkitect.telemetry.trace import TraceConfig, setup_tracing
from arkitect.utils.context import (
    set_account_id,
    set_resource_id,
    set_resource_type,
)

from ..runner import get_default_client_configs


def setup_environment(
    trace_on: bool = True, trace_config: Optional[TraceConfig] = None
) -> None:
    set_resource_type(os.getenv("RESOURCE_TYPE") or "")
    set_resource_id(os.getenv("RESOURCE_ID") or "")
    set_account_id(os.getenv("ACCOUNT_ID") or "")

    setup_tracing(
        endpoint=os.getenv("TRACE_ENDPOINT"),
        trace_on=trace_on,
        trace_config=trace_config,
    )


def initialize(
    context: Any,
    clients: Optional[Dict[str, Tuple[Type[Client], Any]]] = None,
    trace_on: bool = True,
    trace_config: Optional[TraceConfig] = None,
) -> None:
    get_client_pool(clients or get_default_client_configs())

    setup_environment(trace_on, trace_config)



# Copyright (c) 2025 Bytedance Ltd. and/or its affiliates
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ..manifest import ParameterTypeEnum, ToolManifest, ToolParameter


class LinkReader(ToolManifest):
    def __init__(self) -> None:
        super().__init__(
            action_name="LinkReader",
            tool_name="LinkReader",
            description="当你需要获取网页、pdf、抖音视频内容时，使用此工具。"
            + "可以获取url链接下的标题和内容。\n\n"
            + 'examples: {"url_list":["abc.com", "xyz.com"]}',
            parameters=[
                ToolParameter(
                    name="url_list",
                    description="需要解析网页链接,最多3个,以列表返回",
                    param_type=ParameterTypeEnum.ARRAY,
                    items=[
                        ToolParameter(
                            param_type=ParameterTypeEnum.STRING,
                            required=True,
                        )
                    ],
                    required=True,
                )
            ],
        )



# Copyright (c) 2025 Bytedance Ltd. and/or its affiliates
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from enum import Enum
from typing import Any, Dict, List, Optional, Union

from pydantic import BaseModel, Field
from volcenginesdkarkruntime import AsyncArk

from arkitect.core.client import default_ark_client
from arkitect.core.errors import InvalidParameter
from arkitect.telemetry.trace import task

from .model import ArkToolRequest, ArkToolResponse


class ParameterTypeEnum(str, Enum):
    STRING = "string"
    INT = "integer"
    FLOAT = "float"
    BOOL = "boolean"
    ARRAY = "array"
    OBJECT = "object"


class ToolParameter(BaseModel):
    """
    Represents a parameter for a tool.

    Attributes:
        name : The name of the parameter.
        description : A description of the parameter.
        param_type : The type of the parameter.
        required : Whether the parameter is required. Defaults to False.
    """

    param_type: ParameterTypeEnum
    name: str = ""
    description: str = ""
    required: bool = False
    items: List[ToolParameter] = Field(default_factory=list)
    enum: Optional[List[Union[int, bool, str]]] = None

    def validator(self) -> None:
        """
        Validates the parameter.
        """
        pass

    def manifest(self) -> Dict[str, Any]:
        """
        Generates a manifest dictionary for the parameter.
        """
        pm: Dict[str, Any] = {
            "description": self.description,
            "type": self.param_type.value,
        }

        if self.enum:
            pm["enum"] = self.enum

        if len(self.items) > 0 and self.param_type == ParameterTypeEnum.ARRAY:
            pm["items"] = self.items[0].manifest()

        if len(self.items) > 0 and self.param_type == ParameterTypeEnum.OBJECT:
            pm["properties"] = {item.name: item.manifest() for item in self.items}

        return pm


class ToolManifest(BaseModel):
    """
    Represents a manifest for a tool.

    Attributes:
        action_name : The name of the action associated with the tool.
        tool_name : The name of the tool.
        description : A description of the tool. Defaults to None.
        parameters : A list of parameters for the tool.
        manifest_field : A dictionary representing the manifest field.
        client : An instance of the AsyncArk client. Defaults to the default Ark client.
    """

    action_name: str
    tool_name: str
    description: Optional[str] = None
    parameters: List[ToolParameter]

    manifest_field: Dict[str, Any] = Field(default_factory=dict)
    client: AsyncArk = Field(default_factory=default_ark_client)

    class Config:
        """
        Configuration class for the ToolManifest model.
        """

        arbitrary_types_allowed = True

    def __init__(
        self,
        action_name: str,
        tool_name: str,
        description: Optional[str],
        parameters: List[ToolParameter] = [],
        **kwargs: Any,
    ) -> None:
        super().__init__(
            action_name=action_name,
            tool_name=tool_name,
            description=description,
            parameters=parameters,
            **kwargs,
        )

    @classmethod
    def from_manifest(cls, manifest: Dict[str, Any]) -> "ToolManifest":
        """
        Creates a ToolManifest instance from a dictionary.
        This method parses the provided dictionary to create a ToolManifest instance.
        """
        # parse parameters from manifest
        if not manifest:
            raise InvalidParameter("Invalid manifest: empty")

        tm = ToolManifest(
            action_name="", tool_name="", description=manifest.get("description", "")
        )
        if manifest.get("name", "").find("/") > -1:
            parts = manifest.get("name", "").split("/")
            tm.action_name = parts[0]
            tm.tool_name = parts[1]

        else:
            raise InvalidParameter("Invalid manifest: invalid name")

        properties: Dict[str, Any] = manifest.get("parameters", {}).get(
            "properties", {}
        )
        required: List[str] = manifest.get("parameters", {}).get("required", [])
        for name, prop in properties.items():
            param = ToolParameter(**prop)
            param.required = name in required
            tm.parameters.append(param)

        return tm

    def manifest(self) -> Dict[str, Any]:
        """
        Generates a manifest dictionary for the tool.
        """
        if not self.manifest_field:
            self.manifest_field = {
                "name": self.action_name + "/" + self.tool_name,
                "description": self.description or "",
                "parameters": {
                    "properties": {
                        parameter.name: parameter.manifest()
                        for parameter in self.parameters
                    },
                    "required": [
                        parameter.name
                        for parameter in self.parameters
                        if parameter.required
                    ],
                    "type": "object",
                },
            }
        return self.manifest_field

    @task()
    async def executor(
        self, parameters: Optional[Dict[str, Any]] = None, **kwargs: Any
    ) -> ArkToolResponse:
        parameter = ArkToolRequest(
            action_name=self.action_name,
            tool_name=self.tool_name,
            parameters=parameters or {},
        )
        response = await self.client.post(
            path="/tools/execute",
            body=parameter.model_dump(),
            cast_to=ArkToolResponse,
        )
        return ArkToolResponse(**response)

    @property
    def name(self) -> str:
        """
        Returns the full name of the tool.
        """
        return self.action_name + "/" + self.tool_name



# Copyright (c) 2025 Bytedance Ltd. and/or its affiliates
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from typing import Optional, Union

from pydantic import ValidationError
from volcenginesdkarkruntime._exceptions import (
    ArkAPITimeoutError,
    ArkRateLimitError,
)

from arkitect.core.errors.errorcode import ErrorCode
from arkitect.telemetry.logger import ERROR
from arkitect.utils import context

from .errorcode import ArkError


class APIException(Exception):
    def __init__(
        self,
        message: str,
        code: Union[ErrorCode, str],
        parameter: Optional[str] = None,
        http_code: Optional[int] = 500,
        error_type: Optional[str] = "InternalServerError",
    ):
        super().__init__(message)
        self.message = f"{message} Request id: {context.get_reqid()}"
        self.resource_id = context.get_resource_id()
        self.account_id = context.get_account_id() or ""
        self.code = code.value if isinstance(code, ErrorCode) else str(code)
        self.http_code = (
            code.http_code if isinstance(code, ErrorCode) else (http_code or 500)
        )
        self.type = (
            code.error_type
            if isinstance(code, ErrorCode)
            else (error_type or "InternalServerError")
        )
        self.parameter = parameter

    def __str__(self) -> str:
        return (
            "Detailed exception information is listed below.\n"
            + "account_id: {}\n"
            + "resource_id: {}\n"
            + "code: {}\n"
            + "message: {}"
        ).format(self.account_id, self.resource_id, self.code, self.message)

    def to_error(self) -> ArkError:
        return ArkError(
            code=self.code, message=self.message, param=self.parameter, type=self.type
        )


class InternalServiceError(APIException):
    def __init__(self, message: str):
        ERROR(f"[Internal Error]: {message}")
        super().__init__(
            message=ErrorCode.InternalServiceError.message,
            code=ErrorCode.InternalServiceError,
        )


class InvalidParameter(APIException):
    def __init__(self, parameter: str, cause: Optional[str] = None):
        message = ErrorCode.InvalidParameter.message.format(parameter=parameter)
        if cause:
            message = f"{message}.{cause}"
        super().__init__(
            message=message, code=ErrorCode.InvalidParameter, parameter=parameter
        )


class MissingParameter(APIException):
    def __init__(self, parameter: Optional[str] = None):
        message = (
            f"{ErrorCode.MissingParameter.message}:{parameter}"
            if parameter
            else str(ErrorCode.MissingParameter.message)
        )
        super().__init__(
            message=message, code=ErrorCode.MissingParameter, parameter=parameter
        )


class ResourceNotFound(APIException):
    def __init__(self, resource_type: Optional[str] = None):
        message = (
            f"{ErrorCode.ResourceNotFound.message}:{resource_type}"
            if resource_type
            else str(ErrorCode.ResourceNotFound.message)
        )
        super().__init__(
            message,
            code=ErrorCode.ResourceNotFound,
            parameter=resource_type,
        )


class RateLimitExceeded(APIException):
    def __init__(self, resource_type: str):
        message = ErrorCode.RateLimitExceeded.message.format(
            resource_type=resource_type
        )
        super().__init__(
            message,
            code=ErrorCode.RateLimitExceeded,
            parameter=resource_type,
        )


class ServerOverloaded(APIException):
    def __init__(self, service: str):
        message = ErrorCode.ServerOverloaded.message.format(service=service)
        super().__init__(
            message,
            code=ErrorCode.ServerOverloaded,
            parameter=service,
        )


class AuthenticationError(APIException):
    def __init__(self, cause: Optional[str] = None):
        message = ErrorCode.AuthenticationError.message
        if cause:
            message = f"{message} {cause}"
        super().__init__(
            message=message,
            code=ErrorCode.AuthenticationError,
        )


class AccessDenied(APIException):
    def __init__(self, cause: Optional[str] = None):
        message = ErrorCode.AccessDenied.message
        if cause:
            message = f"{message}.{cause}"
        super().__init__(
            message,
            code=ErrorCode.AccessDenied,
        )


class QuotaExceeded(APIException):
    def __init__(self, account_id: str, resource_type: str):
        message = ErrorCode.QuotaExceeded.message.format(
            account_id=account_id, resource_type=resource_type
        )
        super().__init__(message, code=ErrorCode.QuotaExceeded, parameter=resource_type)


class SensitiveContentDetected(APIException):
    def __init__(self, message: str):
        super().__init__(message, code=ErrorCode.SensitiveContentDetected)


class AccountOverdueError(APIException):
    def __init__(self, message: str):
        super().__init__(message, code=ErrorCode.AccountOverdueError)


class APITimeoutError(APIException):
    def __init__(self, message: str):
        ERROR(f"[API timeout error]: {message}")
        super().__init__(
            message=ErrorCode.APITimeoutError.message,
            code=ErrorCode.APITimeoutError,
        )


FALLBACK_EXCEPTIONS = (
    APITimeoutError,
    ArkAPITimeoutError,
    RateLimitExceeded,
    ArkRateLimitError,
)


def parse_pydantic_error(
    e: ValidationError,
) -> Union[MissingParameter, InvalidParameter]:
    try:
        err = e.errors()[-1]
        err_type, err_loc = err.get("type", ""), err.get("loc", ("",))
        if err_type == "missing":
            return MissingParameter(str(err_loc[-1]))
        else:
            return InvalidParameter(str(err_loc[-1]))
    except Exception as e:
        logging.error(f"parse pydantic error:{e}")
        return InvalidParameter("request")



# Copyright (c) 2025 Bytedance Ltd. and/or its affiliates
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .manifest import ParameterTypeEnum, ToolManifest, ToolParameter
from .model import ArkToolRequest, ArkToolResponse
from .pool import ToolPool
from .schema import Calculator, LinkReader

__all__ = [
    "ToolManifest",
    "ToolParameter",
    "ParameterTypeEnum",
    "ToolPool",
    "ArkToolRequest",
    "ArkToolResponse",
    "Calculator",
    "LinkReader",
]



# Copyright (c) 2025 Bytedance Ltd. and/or its affiliates
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import struct
from typing import Optional

from pydantic import BaseModel

from arkitect.core.component.tts.constants import (
    DEFAULT_SPEAKER,
    FULL_CLIENT,
    HEADER_SIZE,
    JSON,
    NAMESPACE,
    NO_COMPRESSION,
    PROTOCAL_VERSION,
    WITH_EVENT,
)


class AudioParams(BaseModel):
    format: str = "mp3"
    sample_rate: int = 24000


class ConnectionParams(BaseModel):
    audio_params: AudioParams
    speaker: str = DEFAULT_SPEAKER


class ResponseEvent(BaseModel):
    event: Optional[int] = None
    audio_only: bool = False
    session_finished: bool = False
    session_id: Optional[str] = None
    connection_id: Optional[str] = None
    payload_msg: dict = {}
    payload_size: int = 0

    audio: bytes = b""


def _write_message(
    event: Optional[int],
    payload: str,
    connection_id: Optional[str] = None,
    session_id: Optional[str] = None,
) -> bytes:
    event_bytes = struct.pack(">I", event)
    frame = event_bytes
    if connection_id is not None:
        connection_id_len = struct.pack(">I", len(connection_id))
        frame += connection_id_len + connection_id.encode("utf-8")
    if session_id is not None:
        session_id_len = struct.pack(">I", len(session_id))
        frame += session_id_len + session_id.encode("utf-8")
    payload_bytes = payload.encode("utf-8")
    payload_len_bytes = struct.pack(">I", len(payload_bytes))
    frame += payload_len_bytes + payload_bytes
    return frame


# Message Class
class Message:
    def __init__(
        self,
        message_type: int = FULL_CLIENT,
        type_flag: int = WITH_EVENT,
        session_id: Optional[str] = None,
        event: Optional[int] = None,
        connection_id: Optional[str] = None,
    ):
        self.type_and_flag_bits = message_type << 4 | type_flag
        self.msg_type = None
        self.event: Optional[int] = event
        self.session_id: Optional[str] = session_id
        self.connection_id: Optional[str] = connection_id
        self.sequence = None
        self.error_code = None
        self.payload: Optional[dict] = None

    def type_flag(self) -> int:
        return self.type_and_flag_bits & 0b00001111

    def write_start_connection(self) -> bytes:
        return self._write_header() + _write_message(event=self.event, payload="{}")

    def write_start_tts_session(self) -> bytes:
        return self._write_header() + _write_message(
            event=self.event,
            connection_id=self.connection_id,
            payload=json.dumps(self.payload),
        )

    def write_text_request(self) -> bytes:
        return self._write_header() + _write_message(
            event=self.event,
            session_id=self.session_id,
            payload=json.dumps(self.payload),
        )

    def write_finish_session(self) -> bytes:
        return self._write_header() + _write_message(
            event=self.event,
            session_id=self.session_id,
            payload=json.dumps(self.payload),
        )

    def write_finish_connection(self) -> bytes:
        return self._write_header() + _write_message(
            event=self.event,
            connection_id=self.connection_id,
            payload=json.dumps(self.payload),
        )

    def _write_header(self) -> bytes:
        header = bytearray(
            [
                PROTOCAL_VERSION << 4 | HEADER_SIZE,
                self.type_and_flag_bits,
                JSON << 4 | NO_COMPRESSION,
            ]
        )
        padding_size = HEADER_SIZE * 4 - len(header)
        if padding_size > 0:
            header.extend([0] * padding_size)
        return header


class TextRequest(BaseModel):
    text: str
    finished: bool


class TTSRequest(BaseModel):
    event: int
    namespace: str = NAMESPACE
    req_params: dict



# Copyright (c) 2025 Bytedance Ltd. and/or its affiliates
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

from arkitect.core.component.llm import BaseChatLanguageModel
from arkitect.core.component.llm.model import ArkMessage

os.environ["ARK_API_KEY"] = "-"


def test_generate_prompts_with_additional_prompts() -> None:
    # Arrange
    messages = [
        ArkMessage(role="user", content="Hello"),
        ArkMessage(role="assistant", content="Hi there!"),
    ]
    additional_prompts = ["Welcome to the chat!", "How can I help you today?"]

    mock_chat_model = BaseChatLanguageModel(endpoint_id="123", messages=messages)
    # Act
    result = mock_chat_model.generate_prompts(
        messages, additional_system_prompts=additional_prompts
    )

    # Assert
    assert len(result) == len(messages) + len(additional_prompts)
    assert all(msg.role == "system" for msg in result[: len(additional_prompts)])
    assert all(msg.role != "system" for msg in result[len(additional_prompts) :])


def test_generate_prompts_without_template():
    # Arrange
    messages = [
        ArkMessage(role="user", content="Hello"),
        ArkMessage(role="assistant", content="Hi there!"),
    ]
    mock_chat_model = BaseChatLanguageModel(endpoint_id="123", messages=messages)
    mock_chat_model.template = None

    # Act
    result = mock_chat_model.generate_prompts(messages)

    # Assert
    assert len(result) == len(messages)


def test_generate_prompts_with_formatting() -> None:
    # Arrange
    messages = [
        ArkMessage(role="system", content="You are a helpful assistant."),
        ArkMessage(role="user", content="Hello"),
        ArkMessage(role="assistant", content="Hi there!"),
    ]
    mock_chat_model = BaseChatLanguageModel(endpoint_id="123", messages=messages)
    additional_prompts = ["Welcome to the chat!", "How can I help you today?"]

    # Act
    result = mock_chat_model.generate_prompts(
        messages, additional_system_prompts=additional_prompts
    )

    # Assert
    assert result
    assert len(result) == len(messages) + len(additional_prompts)
    assert all(msg.role == "system" for msg in result[: len(additional_prompts) + 1])
    assert all(msg.role != "system" for msg in result[len(additional_prompts) + 1 :])



# Copyright (c) 2025 Bytedance Ltd. and/or its affiliates
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List, Optional

from pydantic import BaseModel


class ASRClientConnectRequest(BaseModel):
    pass


class ASRUser(BaseModel):
    uid: Optional[str] = None
    did: Optional[str] = None
    platform: Optional[str] = None
    sdk_version: Optional[str] = None
    app_version: Optional[str] = None


class ASRAudio(BaseModel):
    format: str
    codec: str
    sample_rate: Optional[int] = None
    channel: Optional[int] = None


class ASRCorpus(BaseModel):
    boosting_table_id: Optional[str] = None
    context: Optional[str] = None


class ASRRequest(BaseModel):
    model_name: str
    enable_itn: Optional[bool] = None
    enable_ddc: Optional[bool] = None
    enable_punc: Optional[bool] = None
    corpus: Optional[ASRCorpus] = None


class ASRFullClientRequest(BaseModel):
    user: Optional[ASRUser] = None
    audio: ASRAudio
    request: ASRRequest


class ASRAudioOnlyRequest(BaseModel):
    last_package: bool
    seq: int
    audio: bytes


class ASRAudioInfoRsp(BaseModel):
    duration: Optional[int] = None


class Word(BaseModel):
    blank_duration: Optional[int] = None
    end_time: int
    start_time: int
    text: str


class Utterance(BaseModel):
    definite: bool
    end_time: int
    start_time: int
    text: str
    words: List[Word] = []


class ASRResult(BaseModel):
    text: Optional[str] = ""
    utterances: List[Utterance] = []


class ASRFullServerResponse(BaseModel):
    sequence: Optional[int] = None
    last_package: bool = False
    result: Optional[ASRResult] = None
    audio: Optional[ASRAudioInfoRsp] = None


class ASRServerError(BaseModel):
    code: int
    msg: str

    def __str__(self) -> str:
        return self.msg



# Copyright (c) 2025 Bytedance Ltd. and/or its affiliates
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict, List, Optional


def dict_merge(a: Dict[str, Any], b: Dict[str, Any]) -> Dict[str, Any]:
    """merge two dict recursively and return a new merged dict
    1. if two nodes in the same position are both dicts,
       merge is recursively,
    2. if two nodes in the same position are not both dicts,
       value from the second one overwrites one in first.

    Args:
        a (dict): first dict
        b (dict): second dict

    Returns:
        dict: a new merged dict
    """
    merged = dict()
    for k in set(a.keys()).union(b.keys()):
        if (k in a) and (k in b):
            if isinstance(a[k], dict) and isinstance(b[k], dict):
                merged[k] = dict_merge(a[k], b[k])
            else:
                merged[k] = b[k]
        elif k in a:
            merged[k] = a[k]
        else:  # k in b
            merged[k] = b[k]
    return merged


def list_item_merge(
    a: List[Dict[str, Any]], b: List[Dict[str, Any]], unique_key: Optional[str]
) -> List[Dict[str, Any]]:
    """
    merge two list items into a new one
    will use unique_key to identify the item
    """
    if not unique_key:
        return a + b

    merged = []
    unique_dict = {bb.get(unique_key, ""): bb for bb in b}
    for item in a:
        if item.get(unique_key, "") not in unique_dict.keys():
            merged.append(item)

    merged += b

    return merged



# Copyright (c) 2025 Bytedance Ltd. and/or its affiliates
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
from enum import Enum
from typing import (
    Any,
    AsyncGenerator,
    AsyncIterable,
    Generator,
)

from pydantic import BaseModel


def dump_json_str(obj: Any) -> str:
    """
    Converts an object to a JSON string.
    """
    return json.dumps(dump_json(obj), ensure_ascii=False, default=lambda x: str(x))


def dump_json(obj: Any) -> Any:
    if isinstance(obj, dict):
        return {k: dump_json(v) for k, v in obj.items()}
    elif isinstance(obj, (tuple, list)):
        return [dump_json(v) for v in obj]
    elif isinstance(obj, BaseModel):
        return obj.model_dump(exclude_unset=True, exclude_none=True)
    elif isinstance(obj, (AsyncGenerator, Generator, AsyncIterable)):
        return str(obj)
    else:
        return obj


_MAX_DEPTH = 10


def dump_json_str_truncate(obj: Any, string_length_limit: int) -> str:
    """
    Only for trace
    Truncate all strings to prevent excessively
    long input/output (e.g., base64 images)
    """
    return json.dumps(
        dump_json_truncate(obj, string_length_limit),
        ensure_ascii=False,
        default=lambda x: str(x),
    )


def dump_json_truncate(obj: Any, string_length_limit: int, depth: int = 0) -> Any:
    if depth > _MAX_DEPTH:  # for safety
        return "max recursion depth exceeded"
    if isinstance(obj, dict):
        result_dict = {}
        for k, v in obj.items():
            value = dump_json_truncate(v, string_length_limit, depth + 1)
            if value is not None:
                result_dict[k] = value
        return result_dict
    elif isinstance(obj, Enum):
        return obj.value
    elif isinstance(obj, (AsyncGenerator, Generator, AsyncIterable)):
        return str(obj)
    elif isinstance(obj, (list, tuple)):
        return [
            dump_json_truncate(item, string_length_limit, depth + 1) for item in obj
        ]
    elif isinstance(obj, str):
        return obj[:string_length_limit]
    elif isinstance(obj, BaseModel) and hasattr(obj, "__dict__"):
        result_dict = {}
        for k, v in obj.__dict__.items():
            value = dump_json_truncate(v, string_length_limit, depth + 1)
            if value is not None:
                result_dict[k] = value
        return result_dict
    else:
        return obj



# Copyright (c) 2025 Bytedance Ltd. and/or its affiliates
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import importlib
from typing import AsyncIterable, Callable

from .model import RequestType, ResponseType


def load_function(
    module_name: str, func_name: str
) -> Callable[[RequestType], AsyncIterable[ResponseType]]:
    """
    Loads a function from a specified module.
    """
    package = importlib.import_module(module_name)
    module = getattr(package, func_name)
    return module
